
from collections import OrderedDict, deque
import heapq

from .utils import safe_self_execute
from .errors import LoadInitFailureError
from .packing import Packable
from .transactions import Transactionable
from .hashing import Hashable

class Container(Transactionable, Packable, Hashable):
	pass


# keys must be primitives, values can be primitives or Packable instances/subclasses
class tdict(Container, OrderedDict):
	'''
	Humpack dictionary, replaces the standard dict
	Has all the same functionality of a dict, plus being Transactionable or Packable
	'''
	def __new__(cls, *args, **kwargs):
		
		self = super().__new__(cls)
		
		self.__dict__['_data'] = OrderedDict()
		self.__dict__['_shadow'] = None
		
		return self
	
	def __init__(self, *args, **kwargs):
		super().__init__()
		self.__dict__['_data'] = OrderedDict(*args, **kwargs)
	
	def in_transaction(self):
		return self._shadow is not None
	
	def begin(self):
		if self.in_transaction():
			return
			self.commit()  # partial transactions are committed
		
		self._shadow = self._data
		self._data = self._data.copy()
		
		for child in self.values():  # if keys could be Transactionable instances: chain(self.keys(), self.values())
			if isinstance(child, Transactionable):
				child.begin()
	
	def commit(self):
		if not self.in_transaction():
			return
		
		self._shadow = None
		for child in self.values():  # if keys could be Transactionable instances: chain(self.keys(), self.values())
			if isinstance(child, Transactionable):
				child.commit()
	
	def abort(self):
		if not self.in_transaction():
			return
		
		self._data = self._shadow
		self._shadow = None
		for child in self.values():  # if keys could be Transactionable instances: chain(self.keys(), self.values())
			if isinstance(child, Transactionable):
				child.abort()
	
	def todict(self):
		return {k:v for k,v in self.items()}
	
	def update(self, other):
		self._data.update(other)
	
	def fromkeys(self, keys, value=None):
		self._data.fromkeys(keys, value)
	
	def clear(self):
		self._data.clear()
	
	def copy(self):
		copy = type(self)()
		copy._data = self._data.copy()
		if self._shadow is not None:
			copy._shadow = self._shadow.copy()
		return copy
	
	def __len__(self):
		return len(self._data)
	
	def __hash__(self):
		return id(self)
	
	def __eq__(self, other):
		return id(self) == id(other)
	
	def __contains__(self, item):
		return self._data.__contains__(item)
	
	def __reversed__(self):
		return self._data.__reversed__()
	
	def __iter__(self):
		return iter(self._data)
	
	def keys(self):
		return self._data.keys()
	
	def values(self):
		return self._data.values()
	
	def items(self):
		return self._data.items()
	
	def pop(self, key):
		return self._data.pop(key)
	
	def popitem(self):
		return self._data.popitem()
	
	def move_to_end(self, key, last=True):
		self._data.move_to_end(key, last)
	
	def __pack__(self, pack_member):
		
		data = {}
		
		data['_pairs'] = {}
		data['_order'] = []
		
		for key, value in self.items():
			k, v = pack_member(key, force_str=True), pack_member(value)
			data['_pairs'][k] = v
			data['_order'].append(k)
		
		if self.in_transaction():  # TODO: maybe write warning about saving in the middle of a transaction
			
			data['_shadow_pairs'] = {}
			data['_shadow_order'] = []
			
			for key, value in self._shadow.items():
				k, v = pack_member(key, force_str=True), pack_member(value)
				data['_shadow_pairs'][k] = v
				data['_shadow_order'].append(k)
		
		return data
	
	def __unpack__(self, data, o, unpack_member):
		
		# TODO: write warning about overwriting state - which can't be aborted
		# if self.in_transaction():
		# 	pass
		
		self.abort()
		self._data.clear()
		for key in data['_order']:
			self._data[unpack_member(key)] = unpack_member(data['_pairs'][key])
		
		if '_shadow_pairs' in data:  # TODO: maybe write warning about loading into a partially completed transaction
			self._shadow = OrderedDict()
			for key in data['_shadow_order']:
				self._shadow[unpack_member(key)] = unpack_member(data['_shadow_pairs'][key])
		
		return self
	
	def get(self, k, *args, **kwargs):
		return self._data.get(k, *args, **kwargs)
	
	def setdefault(self, key, default=None):
		self._data.setdefault(key, default)
	
	def __getitem__(self, item):
		return self._data[item]
	
	def __setitem__(self, key, value):  # TODO: write warning if key is not a primitive, subclass of Packable, or instance of Packable
		self._data[key] = value
	
	def __delitem__(self, key):
		del self._data[key]
	
	
	def __str__(self, default='{...}'):
		return safe_self_execute(self, lambda: 't{}{}{}'.format('{', ', '.join(str(key) for key in iter(self)), '}'),
		                         default=default, flag='self printed flag')
	
	def __repr__(self, default='{...}'):
		return safe_self_execute(self, lambda: 't{}{}{}'.format('{', ', '.join(
			('{}:{}'.format(repr(key), repr(value)) for key, value in self.items())), '}'),
		                         default=default, flag='self printed flag')

class adict(tdict):

	def __getattr__(self, item):
		if item in self.__dict__:
			return super().__getattribute__(item)
		return self.__getitem__(item)

	def __setattr__(self, key, value):
		if key in self.__dict__:
			return super().__setattr__(key, value)
		return self.__setitem__(key, value)

	def __delattr__(self, item):
		if item in self.__dict__:
			# raise Exception('{} cannot be deleted'.format(item))
			return super().__delattr__(item)
		return self.__delitem__(item)

class tlist(Container, list):
	'''
	Humpack list, replaces the standard list
	Has all the same functionality of a list, plus being Transactionable or Packable
	'''
	
	def __new__(cls, *args, **kwargs):
		
		self = super().__new__(cls)
		
		self._data = []
		self._shadow = None
		
		return self
	
	def __init__(self, *args, **kwargs):
		super().__init__()
		self._data = list(*args, **kwargs)
	
	def in_transaction(self):
		return self._shadow is not None
	
	def begin(self):
		if self.in_transaction():
			return
			self.commit()  # partial transactions are committed
		
		self._shadow = self._data
		self._data = self._data.copy()
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.begin()
	
	def commit(self):
		if not self.in_transaction():
			return
		
		self._shadow = None
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.commit()
	
	def abort(self):
		if not self.in_transaction():
			return
		
		self._data = self._shadow
		self._shadow = None
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.abort()
	
	def tolist(self):
		return [x for x in self]
	
	def copy(self):
		copy = type(self)()
		copy._data = self._data.copy()
		if self._shadow is not None:
			copy._shadow = self._shadow.copy()
		return copy
	
	def __pack__(self, pack_member):
		state = {}
		state['_entries'] = [pack_member(elm) for elm in iter(self)]
		if self.in_transaction():  # TODO: maybe write warning about saving in the middle of a transaction
			state['_shadow'] = [pack_member(elm) for elm in self._shadow]
		return state
	
	def __unpack__(self, state, o, unpack_member):
		
		# TODO: write warning about overwriting state - which can't be aborted
		# if self.in_transaction():
		# 	pass
		
		self._data.extend(unpack_member(elm) for elm in state['_entries'])
		if '_shadow' in state:  # TODO: maybe write warning about loading into a partially completed transaction
			self._shadow = [unpack_member(elm) for elm in state['_shadow']]
	
	def __getitem__(self, item):
		if isinstance(item, slice):
			return tlist(self._data[item])
		return self._data[item]
	
	def __setitem__(self, key, value):
		self._data[key] = value
	
	def __delitem__(self, idx):
		del self._data[idx]
	
	def __hash__(self):
		return id(self)
	
	def __eq__(self, other):
		return id(self) == id(other)
	
	def count(self, object):
		return self._data.count(object)
	
	def append(self, item):
		return self._data.append(item)
	
	def __contains__(self, item):
		return self._data.__contains__(item)
	
	def extend(self, iterable):
		return self._data.extend(iterable)
	
	def insert(self, index, object):
		self._data.insert(index, object)
	
	def remove(self, value):
		self._data.remove(value)
	
	def __iter__(self):
		return iter(self._data)
	
	def __reversed__(self):
		return self._data.__reversed__()
	
	def reverse(self):
		self._data.reverse()
	
	def pop(self, index=None):
		if index is None:
			return self._data.pop()
		return self._data.pop(index)
	
	def __len__(self):
		return len(self._data)
	
	def clear(self):
		self._data.clear()
	
	def sort(self, key=None, reverse=False):
		self._data.sort(key=key, reverse=reverse)
	
	def index(self, object, start=None, stop=None):
		self._data.index(object, start, stop)
	
	def __mul__(self, other):
		return tlist(self._data.__mul__(other))
	
	def __rmul__(self, other):
		return tlist(self._data.__rmul__(other))
	
	def __add__(self, other):
		out = self.copy()
		out.extend(other)
		return out
	
	def __iadd__(self, other):
		self._data.__iadd__(other)
	
	def __imul__(self, other):
		self._data.__imul__(other)
	
	def __str__(self, default='[...]'):
		return safe_self_execute(self, lambda: 't[{}]'.format(', '.join(map(str, self))),
		                         default=default, flag='self printed flag')
	
	def __repr__(self, default='[...]'):
		return safe_self_execute(self, lambda: 't[{}]'.format(', '.join(map(repr, self))),
		                         default=default, flag='self printed flag')

class tset(Container, set):
	'''
	Humpack set, replaces the standard set
	Has all the same functionality of a set, plus being Transactionable or Packable
	'''
	
	def __new__(cls, *args, **kwargs):
		
		self = super().__new__(cls)
		
		self._data = OrderedDict()
		self._shadow = None
		
		return self
	
	def __init__(self, iterable=[]):
		super().__init__()
		for x in iterable:
			self.add(x)
		self._shadow = None
	
	def in_transaction(self):
		return self._shadow is not None
	
	def begin(self):
		if self.in_transaction():
			return
			self.commit()  # partial transactions are committed
		
		self._shadow = self._data
		self._data = self._data.copy()
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.begin()
	
	def commit(self):
		if not self.in_transaction():
			return
		
		self._shadow = None
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.commit()
	
	def abort(self):
		if not self.in_transaction():
			return
		
		self._data = self._shadow
		self._shadow = None
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.abort()
	
	def toset(self):
		return {x for x in self}
	
	def copy(self):
		copy = type(self)()
		copy._data = self._data.copy()
		if self._shadow is not None:
			copy._shadow = self._shadow.copy()
		return copy
	
	def __pack__(self, pack_member):
		state = {}
		state['_elements'] = [pack_member(elm) for elm in iter(self)]
		if self.in_transaction():
			state['_shadow'] = [pack_member(elm) for elm in self._shadow]
		return state
	
	def __unpack__(self, data, o, unpack_member):
		
		# TODO: write warning about overwriting state - which can't be aborted
		# if self.in_transaction():
		# 	pass
		
		self.update(unpack_member(elm) for elm in data['_elements'])
		
		if '_shadow' in data:  # TODO: maybe write warning about loading into a partially completed transaction
			self._shadow = OrderedDict()
			for elm in data['_shadow']:
				self._shadow[unpack_member(elm)] = None
	
	def __hash__(self):
		return id(self)
	
	def __eq__(self, other):
		return id(self) == id(other)
	
	def __and__(self, other):
		copy = self.copy()
		for x in self:
			if x in other:
				copy.add(x)
			else:
				copy.remove(x)
		return copy
	
	def __or__(self, other):
		copy = self.copy()
		copy.update(other)
		return copy
	
	def __xor__(self, other):
		copy = self.copy()
		for x in list(other):
			if x in other:
				copy.add(x)
			else:
				copy.remove(x)
		return copy
	
	def __sub__(self, other):
		copy = self.copy()
		for x in other:
			copy.discard(x)
		return copy
	
	def __rand__(self, other):
		return self & other
	
	def __ror__(self, other):
		return self | other
	
	def __rxor__(self, other):
		return self ^ other
	
	def __rsub__(self, other):
		copy = other.copy()
		for x in self:
			copy.discard(x)
		return copy
	
	def difference_update(self, other):
		self -= other
	
	def intersection_update(self, other):
		self &= other
	
	def union_update(self, other):
		self |= other
	
	def symmetric_difference_update(self, other):
		self ^= other
	
	def symmetric_difference(self, other):
		return self ^ other
	
	def union(self, other):
		return self | other
	
	def intersection(self, other):
		return self & other
	
	def difference(self, other):
		return self - other
	
	def issubset(self, other):
		for x in self:
			if x not in other:
				return False
		return True
	
	def issuperset(self, other):
		for x in other:
			if x not in self:
				return False
		return True
	
	def isdisjoint(self, other):
		return not self.issubset(other) and not self.issuperset(other)
	
	def __iand__(self, other):
		for x in list(self):
			if x not in other:
				self.remove(x)
	
	def __ior__(self, other):
		self.update(other)
	
	def __ixor__(self, other):
		for x in other:
			if x in self:
				self.remove(x)
			else:
				self.add(x)
	
	def __isub__(self, other):
		for x in other:
			if x in self:
				self.remove(x)
	
	def pop(self):
		return self._data.popitem()[0]
	
	def remove(self, item):
		del self._data[item]
	
	def discard(self, item):
		if item in self._data:
			self.remove(item)
	
	def __contains__(self, item):
		return self._data.__contains__(item)
	
	def __len__(self):
		return len(self._data)
	
	def __iter__(self):
		return iter(self._data)
	
	def clear(self):
		return self._data.clear()
	
	def update(self, other):
		for x in other:
			self.add(x)
	
	def add(self, item):
		self._data[item] = None
	
	def __str__(self, default='{...}'):
		return safe_self_execute(self, lambda: 't{}{}{}'.format('{', ', '.join(map(str, self)), '}'),
		                         default=default, flag='self printed flag')
	
	def __repr__(self, default='{...}'):
		return safe_self_execute(self, lambda: 't{}{}{}'.format('{', ', '.join(map(repr, self)), '}'),
		                         default=default, flag='self printed flag')


class tdeque(Container, deque):
	'''
	Humpack queue, replaces the standard deque
	Has all the same functionality of a set, plus being Transactionable or Packable
	'''
	
	def __new__(cls, *args, **kwargs):
		
		self = super().__new__(cls)
		
		self._data = deque()
		self._shadow = None
		
		return self
	
	def __init__(self, *args, **kwargs):
		super().__init__()
		self._data = deque(*args, **kwargs)
	
	def in_transaction(self):
		return self._shadow is not None
	
	def begin(self):
		if self.in_transaction():
			return
			self.commit()  # partial transactions are committed
		
		self._shadow = self._data
		self._data = self._data.copy()
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.begin()
	
	def commit(self):
		if not self.in_transaction():
			return
		
		self._shadow = None
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.commit()
	
	def abort(self):
		if not self.in_transaction():
			return
		
		self._data = self._shadow
		self._shadow = None
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.abort()
	
	def copy(self):
		copy = type(self)()
		copy._data = self._data.copy()
		if self._shadow is not None:
			copy._shadow = self._shadow.copy()
		return copy
	
	def __pack__(self, pack_member):
		state = {}
		state['_entries'] = [pack_member(elm) for elm in iter(self)]
		if self.in_transaction():  # TODO: maybe write warning about saving in the middle of a transaction
			state['_shadow'] = [pack_member(elm) for elm in self._shadow]
		return state
	
	def __unpack__(self, state, o, unpack_member):
		
		# TODO: write warning about overwriting state - which can't be aborted
		# if self.in_transaction():
		# 	pass
		
		self._data.extend(unpack_member(elm) for elm in state['_entries'])
		if '_shadow' in state:  # TODO: maybe write warning about loading into a partially completed transaction
			self._shadow = [unpack_member(elm) for elm in state['_shadow']]
	
	def __getitem__(self, item):
		if isinstance(item, slice):
			return tdeque(self._data[item])
		return self._data[item]
	
	def __setitem__(self, key, value):
		self._data[key] = value
	
	def __delitem__(self, idx):
		del self._data[idx]
	
	def __hash__(self):
		return id(self)
	
	def __eq__(self, other):
		return id(self) == id(other)
	
	def count(self, object):
		return self._data.count(object)
	
	def append(self, item):
		return self._data.append(item)
	
	def appendleft(self, item):
		return self._data.appendleft(item)
	
	def __contains__(self, item):
		return self._data.__contains__(item)
	
	def extend(self, iterable):
		return self._data.extend(iterable)
	
	def extendleft(self, iterable):
		return self._data.extendleft(iterable)
	
	def insert(self, index, object):
		self._data.insert(index, object)
	
	def remove(self, value):
		self._data.remove(value)
	
	def __iter__(self):
		return iter(self._data)
	
	def __reversed__(self):
		return self._data.__reversed__()
	
	def reverse(self):
		self._data.reverse()
	
	def pop(self):
		return self._data.pop()
		
	def popleft(self):
		return self._data.popleft()
		
	def __len__(self):
		return len(self._data)
	
	def clear(self):
		self._data.clear()
	
	def sort(self, key=None, reverse=False):
		self._data.sort(key, reverse)
	
	def index(self, object, start=None, stop=None):
		self._data.index(object, start, stop)
	
	def rotate(self, n=1):
		return self._data.rotate(n=n)
	
	def __mul__(self, other):
		return tdeque(self._data.__mul__(other))
	
	def __rmul__(self, other):
		return tdeque(self._data.__rmul__(other))
	
	def __add__(self, other):
		out = self.copy()
		out.extend(other)
		return out
	
	def __iadd__(self, other):
		self._data.__iadd__(other)
	
	def __imul__(self, other):
		self._data.__imul__(other)
	
	def __str__(self, default='[...]'):
		return safe_self_execute(self, lambda: 't[{}]'.format(', '.join(map(str, self))),
		                         default=default, flag='self printed flag')
	
	def __repr__(self, default='[...]'):
		return safe_self_execute(self, lambda: 't[{}]'.format(', '.join(map(repr, self))),
		                         default=default, flag='self printed flag')

class tstack(tdeque):
	'''
	Humpack stack
	Has all the same functionality of a deque, except it's a stack (FIFO)
	Also implements Transactionable and Packable
	'''
	
	def pop(self):
		return super().popleft()
	
	def popend(self):
		return super().pop()
	
	def push(self, item):
		return super().appendleft(item)
	
	def push_all(self, items):
		return super().extendleft(reversed(items))
	
	def peek(self, n=0):
		return self[n]


class _theap_iter(object):
	def __init__(self, heap):
		self._heap = heap
	
	def __next__(self):
		if len(self._heap):
			return self._heap.pop()
		raise StopIteration

class theap(Container, object):
	'''
	Humpack heap
	Unordered for adding/removing, ordered when iterating.
	Note that iterating through the heap empties it.
	'''
	
	def __new__(cls, *args, **kwargs):
		
		self = super().__new__(cls)
		
		self._data = []
		self._shadow = None
		
		return self
	
	def __init__(self, *args, **kwargs):
		super().__init__()
		self._data = list(*args, **kwargs)
		heapq.heapify(self._data)
	
	def in_transaction(self):
		return self._shadow is not None
	
	def begin(self):
		if self.in_transaction():
			return
			self.commit()  # partial transactions are committed
		
		self._shadow = self._data
		self._data = self._data.copy()
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.begin()
	
	def commit(self):
		if not self.in_transaction():
			return
		
		self._shadow = None
		
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.commit()
	
	def abort(self):
		if not self.in_transaction():
			return
		
		self._data = self._shadow
		self._shadow = None
		for child in iter(self):
			if isinstance(child, Transactionable):
				child.abort()
	
	def copy(self):
		copy = type(self)()
		copy._data = self._data.copy()
		if self._shadow is not None:
			copy._shadow = self._shadow.copy()
		return copy
	
	def __pack__(self, pack_member):
		state = {}
		state['_entries'] = [pack_member(elm) for elm in iter(self)]
		if self.in_transaction():  # TODO: maybe write warning about saving in the middle of a transaction
			state['_shadow'] = [pack_member(elm) for elm in self._shadow]
		return state
	
	def __unpack__(self, state, o, unpack_member):
		
		# TODO: write warning about overwriting state - which can't be aborted
		# if self.in_transaction():
		# 	pass
		
		self._data.extend(unpack_member(elm) for elm in state['_entries'])
		if '_shadow' in state:  # TODO: maybe write warning about loading into a partially completed transaction
			self._shadow = [unpack_member(elm) for elm in state['_shadow']]
			
	def __iter__(self): # Note: this actually pops entries - iterating through heap will empty it
		return _theap_iter(self.copy())
	
	def __len__(self):
		return len(self._data)
		
	def push(self, *items):
		for item in items:
			heapq.heappush(self._data, item)

	def pop(self, n=None):
		if n is None:
			return heapq.heappop(self._data)
		return tlist(heapq.heappop(self._data) for _ in range(n))
	
	def replace(self, item):
		return heapq.heapreplace(self._data, item)
	
	def pushpop(self, item):
		return heapq.heappushpop(self._data, item)
	
	def __hash__(self):
		return id(self)
	
	def __eq__(self, other):
		return id(self) == id(other)
	
	def __str__(self, default='[...]'):
		return safe_self_execute(self, lambda: 't[{}]'.format(', '.join(map(str, self._data))),
		                         default=default, flag='self printed flag')
	
	def __repr__(self, default='[...]'):
		return safe_self_execute(self, lambda: 't[{}]'.format(', '.join(map(repr, self._data))),
		                         default=default, flag='self printed flag')


def containerify(obj, dtype=tdict):
	'''
	Recursively, convert `obj` from using standard python containers to HumPack containers.

	:param obj: object using python containers (dict, list, set, tuple, etc.)
	:return: deep copy of the object using HumPack containers
	'''
	if isinstance(obj, deque):
		return tdeque(containerify(o, dtype=dtype) for o in obj)
	if isinstance(obj, list):
		return tlist(containerify(o, dtype=dtype) for o in obj)
	if isinstance(obj, set):
		return tset(containerify(o, dtype=dtype) for o in obj)
	if isinstance(obj, tuple):
		return tuple(containerify(o, dtype=dtype) for o in obj)
	if isinstance(obj, dict):
		return dtype({containerify(k, dtype=dtype): containerify(v, dtype=dtype) for k, v in obj.items()})
	
	return obj





# from omnibelt import safe_self_execute

from collections import OrderedDict

def sign(x):
	return 0 if x == 0 else (1 if x > 0 else -1)


def safe_self_execute(obj, fn, default='<<short circuit>>',
                 flag='safe execute flag'):
	
	if flag in obj.__dict__:
		return default  # short circuit
	obj.__dict__[flag] = True
	
	try:
		out = fn()
	finally:
		del obj.__dict__['self printed flag']
	
	return out



def split_dict(items, keys):
	good, bad = OrderedDict(), OrderedDict()
	for k in items:
		if k in keys:
			good[k] = items[k]
		else:
			bad[k] = items[k]
	return good, bad











from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, IO
from pathlib import Path
from . import Exporter, load_txt, save_txt, Packable, save_pack, load_pack, save_json, save_yaml, load_yaml, load_json
from .exporting import ExportManager, Exporter, SelectiveExporter, CollectiveExporter
import dill
import toml


class PickleExport(CollectiveExporter, extensions='.pk'):
	@staticmethod
	def validate_export_obj(obj: Any, **kwargs) -> bool:
		return dill.pickles(obj, **kwargs)

	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager], **kwargs) -> Any:
		return dill.load(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str], src: Type[ExportManager], **kwargs) -> Path:
		dill.dump(payload, path, **kwargs)
		return path


class PackedExport(CollectiveExporter, extensions='.pkd'):
	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager], **kwargs) -> Any:
		return load_pack(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str],
	                    src: Type[ExportManager], **kwargs) -> Optional[Path]:
		save_pack(payload, path, **kwargs)
		return path



class JsonExport(CollectiveExporter, extensions='.json'):
	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager], **kwargs) -> Any:
		return load_json(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str],
	                    src: Type[ExportManager], **kwargs) -> Optional[Path]:
		save_json(payload, path, **kwargs)
		return path


class YamlExport(CollectiveExporter, extensions=['.yaml', '.yml']):
	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager], **kwargs) -> Any:
		return load_yaml(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str],
	                    src: Type[ExportManager], **kwargs) -> Optional[Path]:
		save_yaml(payload, path, **kwargs)
		return path


class TomlExport(CollectiveExporter, extensions=['.toml', '.tml']):
	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager], **kwargs) -> Any:
		return toml.load(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str],
	                    src: Type[ExportManager], **kwargs) -> Optional[Path]:
		toml.dump(payload, path.open('w'), **kwargs)
		return path



class StrExport(SelectiveExporter, types=str, extensions='.txt'):
	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager]) -> str:
		return load_txt(path)

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str],
	                    src: Type[ExportManager], **kwargs) -> Optional[Path]:
		save_txt(payload, path)
		return path


class IntExport(SelectiveExporter, types=int, extensions='.int'):
	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager]) -> int:
		return int(load_txt(path))

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str],
	                    src: Type[ExportManager], **kwargs) -> Optional[Path]:
		return save_txt(str(payload), path)


class FloatExport(SelectiveExporter, types=float, extensions='.float'):
	@staticmethod
	def _load_export(path: Union[Path, str], src: Type[ExportManager]) -> float:
		return float(load_txt(path))

	@staticmethod
	def _export_payload(payload: Any, path: Union[Path, str],
	                    src: Type[ExportManager], **kwargs) -> Optional[Path]:
		return save_txt(str(payload), path)







import types
from typing import Callable, Any
import inspect


primitives = (str, int, float, bool, type(None))



class unspecified_argument:
	@staticmethod
	def __repr__():
		return '<unspecified>'


class agnostic:
	def __init__(self, fn=None, **kwargs):
		super().__init__(**kwargs)
		self.fn = fn

	def __get__(self, instance, owner=None):
		assert owner is not None, f'owner is missing: {self.fn} ({instance})'
		if instance is None:
			instance = owner
		return self.fn.__get__(instance, owner)


class agnosticproperty(property):
	def __get__(self, instance, owner=None):
		assert owner is not None, f'owner is missing: {self.fget} ({instance})'
		if instance is None:
			instance = owner
		return super().__get__(instance, owner)

	def getter(self, fget: Callable[[Any], Any]) -> property:
		return type(self)(fget, self.fset, self.fdel, self.__doc__)

	def setter(self, fset: Callable[[Any, Any], None]) -> 'agnosticproperty':
		return type(self)(self.fget, fset, self.fdel, self.__doc__)
	
	def deleter(self, fdel: Callable[[Any], None]) -> 'agnosticproperty':
		return type(self)(self.fget, self.fset, fdel, self.__doc__)


agnosticmethod = agnostic

# class agnosticmethod:
# 	def __init__(self, fn):
# 		self.fn = fn
#
#
# 	def __get__(self, obj, cls):
# 		# if inspect.isgeneratorfunction(self.fn):
# 		# 	return types.GeneratorType(self.fn, obj)
# 		return types.MethodType(self.fn, cls if obj is None else obj)


# class agnosticproperty:
#     def __init__(self, fget, fset=None):
#         self.fget = fget
#         self.fset = fset
#
#     def __get__(self, obj, klass=None):
#         if klass is None:
#             klass = type(obj)
#         return self.fget.__get__(obj, klass)()
#
#     def __set__(self, obj, value):
#         if not self.fset:
#             raise AttributeError("can't set attribute")
#         type_ = type(obj)
#         return self.fset.__get__(obj, type_)(value)
#
#     def setter(self, func):
#         if not isinstance(func, agnosticmethod):
#             func = agnosticmethod(func)
#         self.fset = func
#         return self



def duplicate_func(f, cls=None, name=None):
	'''
	Adapted from Aaron Hall's post here:
	https://stackoverflow.com/questions/6527633
	'''

	closure = []
	if f.__closure__ is not None:
		closure = list(f.__closure__)

		if cls is not None:
			cls_cell_idx = f.__code__.co_freevars.index("__class__")
			closure[cls_cell_idx] = types.CellType(cls)

	if name is None:
		name = f.__name__

	new = types.FunctionType(
		f.__code__,
		f.__globals__,
		name,
		f.__defaults__,
		tuple(closure)
	)

	new.__qualname__ = f"{cls.__name__}.{new.__name__}"
	new.__dict__.update(f.__dict__)

	return new



def join_classes(*bases, name=None, data=None):
	if data is None:
		data = {}
	if name is None:
		name = '_'.join(cls.__name__ for cls in bases)
	return type(name, bases, data)


class ClassAttr:
	def duplicate(self):
		clone = _Blank()
		clone.__class__ = self.__class__
		clone.__dict__.update(self.__dict__)
		return clone



def duplicate_class(cls, name=None, chain=False, data=None):
	if name is None:
		name = cls.__name__

	if chain:
		parents = (cls,)
		data = {}
	else:
		parents = cls.__bases__
		data = dict(cls.__dict__)

	new = join_classes(*parents, name=name, data=data)
	for key, attr in new.__dict__.items():
		if isinstance(attr, ClassAttr):
			setattr(new, key, attr.duplicate())
		elif isinstance(attr, types.FunctionType):
			setattr(new, key, duplicate_func(attr, cls=new))

	return new



class _Blank: pass
def duplicate_instance(obj):
	# slots = getattr(obj.__class__, '__slots__', None)
	# if slots is None:
	# 	class _Blank():
	# 		pass
	# else:
	# 	class _Blank():
	# 		__slots__ = slots
	new = _Blank()
	new.__dict__.update(obj.__dict__)
	new.__class__ = obj.__class__
	return new



class conditional_method(ClassAttr):
	def __init__(self, fn=None):
		self.fn = fn

	def condition(self, instance):
		raise NotImplementedError

	def set_owner(self, owner):
		self.owner = owner

	def set_base(self, base):
		self.base = base

	def _build_method(self, fn, instance, owner):
		return types.MethodType(duplicate_func(fn, cls=owner), instance)

	def __get__(self, instance, owner):
		# print(f"returned from descriptor object {instance} {owner}")

		if instance is None:
			return self

		if not self.condition(instance):
			raise AttributeError(f'Condition of wrapped method failed: {self.fn.__name__}')
		meth = self._build_method(self.fn, instance, getattr(self, 'base', owner))
		return meth


	def __set__(self, instance, value):
		print(f"set in descriptor object {instance} {value}")
		self.fn = value


	def __call__(self, fn):
		self.fn = fn
		return self



class lambda_conditional_method(conditional_method):
	def __init__(self, condition):
		self._condition = condition


	def condition(self, instance):
		return self._condition(instance)



_class_subs = {}
def _gen_sub_template_name(cls):
	if cls not in _class_subs:
		_class_subs[cls] = 0
	_class_subs[cls] += 1
	return f'{cls.__name__}{_class_subs[cls]}'



def wrap_class(wrapper, cls, name=None, chain=False, data=None):
	if name is None:
		name = _gen_sub_template_name(wrapper)

	sub = duplicate_class(wrapper, name, chain=chain, data=data)
	for attr in sub.__dict__.values():
		if isinstance(attr, conditional_method):
			attr.set_owner(sub)
			attr.set_base(cls)
	return join_classes(sub, cls)



def replace_class(obj, cls, check_conditions=True):
	obj.__class__ = cls
	return obj



def mix_into(wrapper, obj, new_instance=False, chain=False, cls_data=None):
	cls = wrap_class(wrapper, obj.__class__, chain=chain, data=cls_data)
	if new_instance:
		obj = duplicate_instance(obj)
	return replace_class(obj, cls)




# def make_multiplier_of(n):
#     def multiplier(x):
#         print(n)
#         return x * n
#     return multiplier
# f = make_multiplier_of(2)
# c = f.__code__
# c2 = c.replace(co_freevars=('_test', *f.__code__.co_freevars))
# closure = [ types.CellType('!')]
# if f.__closure__ is not None:
#     closure.extend(f.__closure__)
# f2 = types.FunctionType(
#     c2,
#     f.__globals__,
#     'f2',
#     f.__defaults__,
#     tuple(closure)
# )






name = 'omnibelt'
long_name = 'omni-belt'

version = '0.6.0'
url = 'https://github.com/felixludos/omni-belt'

description = 'Universal python utilities'

author = 'Felix Leeb'
author_email = 'felixludos.info@gmail.com'

license = 'MIT'

readme = 'README.md'

packages = ['omnibelt']

import os
try:
	with open(os.path.join(os.path.abspath(os.path.dirname(os.path.dirname(__file__))), 'requirements.txt'), 'r') as f:
		install_requires = f.readlines()
except:
	install_requires = ['pyyaml', 'c3linearize', 'wrapt', 'cryptography']
del os





from typing import NamedTuple, Dict, Any, Union, Optional, List, Tuple, Callable, Type, Sequence, Iterable, Iterator
from pathlib import Path
from collections import OrderedDict, namedtuple

from .typing import unspecified_argument
from .logging import get_printer

prt = get_printer(__name__)

class Registry(OrderedDict):
	def new(self, name, obj): # register a new entry
		# if name in self:
		# 	prt.warning(f'Register {self.__class__.__name__} already contains {name}, now overwriting')
		# else:
		# 	prt.debug(f'Registering {name} in {self.__class__.__name__}')
		
		self[name] = obj
		return obj

	class NotFoundError(KeyError): pass


	def find(self, x, default=unspecified_argument):
		if x in self:
			return self[x]
		if default is not unspecified_argument:
			return default
		raise self.NotFoundError(x)


	def is_registered(self, obj):
		for opt in self.values():
			if obj == opt:
				return True
		return False


class Named_Registry(Registry):


	def find_name(self, obj):
		return obj.get_name()


	def is_registered(self, obj):
		name = self.find_name(obj)
		return name in self


# class _Entry:
# 	def __init__(self, **kwargs):
# 		self.__dict__.update(kwargs)


class Entry_Registry(Registry):
	'''
	Automatically wraps data into an "entry" object (namedtuple) which is stored in the registry
	'''
	entry_cls: namedtuple = None
	def __init_subclass__(cls, key_name='name', components=[], required=[]):
		super().__init_subclass__()
		entry_keys = [key_name, *components]
		# cls._entry_keys = entry_keys
		cls.entry_cls: namedtuple = namedtuple(f'{cls.__name__}_Entry', entry_keys)
		cls._required_keys = [key_name, *required]
		cls._key_name = key_name


	def new(self, *args, **kwargs):  # register a new entry
		args = dict(zip(self.entry_cls._fields, args))

		overlap = ', '.join(set(args).intersection(kwargs.keys()))
		if len(overlap):
			raise TypeError(f'{self.__class__.__name__} got multiple values for arguments: {overlap}')

		info = {**args, **kwargs}

		missing = ', '.join(key for key in self._required_keys if key not in info)
		if len(missing):
			raise TypeError(f'{self.__class__.__name__} missing {len(missing)} required keys for entry: {missing}')

		return super().new(info[self._key_name], self.entry_cls(**info))



class InvalidDoubleRegistryError(Exception):
	pass



class Double_Registry(Registry):
	def __init__(self, *args, _sister_registry_object=None, _sister_registry_cls=None, **kwargs):
		if _sister_registry_cls is None:
			_sister_registry_cls = self.__class__
		if _sister_registry_object is None:
			_sister_registry_object = _sister_registry_cls(_sister_registry_object=self)
		
		super().__init__(*args, **kwargs)
		self._sister_registry_object = _sister_registry_object
		
		self._init_sister_registry()


	def _init_sister_registry(self):
		for k, v in self.items():
			self._sister_registry_object.__setitem__(v, k, sync=False)


	def is_known(self, x):
		return x in self or x in self.backward()


	def find(self, x, default=unspecified_argument):
		if x in self:
			return self[x]
		if x in self.backward():
			return self.backward()[x]
		if default is not unspecified_argument:
			return default
		raise self.NotFoundError(x)


	def backward(self):
		return self._sister_registry_object


	def update(self, other, sync=True):
		if sync:
			self._sister_registry_object.update({v:k for k,v in other.items()}, sync=False)
		return super().update(other)


	def __setitem__(self, key, value, sync=True):
		if sync:
			self._sister_registry_object.__setitem__(value, key, sync=False)
		return super().__setitem__(key, value)


	def __delitem__(self, key, sync=True):
		if sync:
			self._sister_registry_object.__delitem__(self[key], sync=False)
		return super().__delitem__(key)



class Entry_Double_Registry(Double_Registry, Entry_Registry):
	
	def __init_subclass__(cls, primary_component='name', sister_component='value', components=[], required=[]):
		super().__init_subclass__(key_name=primary_component, components=[sister_component, *components],
		                          required=[sister_component, *required])
		cls._sister_key_name = sister_component


	def _init_sister_registry(self):
		for k, v in self.items():
			self._sister_registry_object.__setitem__(self._get_sister_entry_key(k, v), v, sync=False)


	@classmethod
	def _get_sister_entry_key(cls, key, value):
		assert isinstance(value, cls.entry_cls)
		if key == getattr(value, cls._key_name):
			return getattr(value, cls._sister_key_name)
		return getattr(value, cls._key_name)


	def get_value(self, key):
		return getattr(self.find(key), self._sister_key_name)


	def update(self, other, sync=True):
		if sync:
			self._sister_registry_object.update({self._get_sister_entry_key(k, v): v
			                                     for k, v in other.items()}, sync=False)
		return super().update(other, sync=False)


	def __setitem__(self, key, value, sync=True):
		if sync:
			self._sister_registry_object.__setitem__(self._get_sister_entry_key(key, value), value, sync=False)
		return super().__setitem__(key, value, sync=False)


	def __delitem__(self, key, sync=True):
		if sync:
			self._sister_registry_object.__delitem__(self._get_sister_entry_key(key, self[key]), sync=False)
		return super().__delitem__(key, sync=False)


	def get_decorator(self, name: Optional[str] = None, defaults: Dict[str, Any] = None):
		if defaults is None:
			defaults = {}
		return type(f'{self.__class__.__name__}_Decorator' if name is None else name,
		            (self.DecoratorBase,), {'_registry': self, '_defaults': defaults})


	class DecoratorBase:
		_registry = None
		_defaults = None
		def __init__(self, *args, **kwargs):

			registry = self._registry

			arg_keys = list(registry.entry_cls._fields)
			# del arg_keys[1]
			args = dict(zip(arg_keys, args))

			overlap = ', '.join({registry._sister_key_name, *args.keys()}.intersection(kwargs))
			if len(overlap):
				raise TypeError(f'{self.__class__.__name__} got multiple values for arguments: {overlap}')

			self.params = {**args, **kwargs}


		# @classmethod
		# def _get_registry(cls):
		# 	return cls._registry


		def _register(self, val, **params):
			registry = self._registry
			key = registry._sister_key_name
			if key not in params:
				params[key] = val
			full = self._defaults.copy()
			full.update(params)
			return registry.new(**params)


		def __call__(self, sister_value):
			self._register(sister_value, **self.params)
			return sister_value


class Path_Registry(Entry_Double_Registry, sister_component='path'):
	def __init_subclass__(cls, sister_component='path', components=[], required=[]):
		super().__init_subclass__(primary_component='name', sister_component=sister_component,
		                          components=components, required=required)

	def new(self, name, path, *args, **kwargs):
		if isinstance(path, str):
			path = Path(path)
		return super().new(name, path, *args, **kwargs)

	def get_path(self, key):
		return self.get_value(key)


class Function_Registry(Entry_Double_Registry, sister_component='fn'):
	def __init_subclass__(cls, sister_component='fn', components=[], required=[]):
		super().__init_subclass__(primary_component='name', sister_component=sister_component,
		                          components=components, required=required)

	def get_function(self, key):
		return self.get_value(key)


class Class_Registry(Entry_Double_Registry, sister_component='cls'):

	def __init_subclass__(cls, sister_component='cls', components=[], required=[]):
		super().__init_subclass__(primary_component='name', sister_component=sister_component,
		                          components=components, required=required)

	def get_class(self, key):
		return self.get_value(key)

	class DecoratorBase(Entry_Double_Registry.DecoratorBase):
		def _register(self, val, name=None, **params):
			if name is None:
				name = val.__name__
			return super()._register(val, name=name, **params)








from datetime import datetime
import time

STD_FMT = "%y%m%d-%H%M%S"

def get_now(fmt=STD_FMT):
	now = datetime.now()
	if fmt is None:
		return now
	return now.strftime(fmt)

def recover_date(raw, fmt=STD_FMT):
	return datetime.strptime(raw, fmt)






# packing

class WrongKeyError(Exception):
	'''Error thrown when the provided key for decryption is incorrect'''
	def __init__(self):
		super().__init__('Unable to decrypt the data because provided hash is invalid')

class LoadInitFailureError(Exception):
	'''Error thrown when an object cannot be recreated'''
	def __init__(self, obj_type):
		super().__init__('An instance of {obj_type} was unable to load (make sure {obj_type}.__init__ doesnt '
		                 'have any required arguments)'.format(obj_type=obj_type))

class ObjectIDReadOnlyError(Exception):
	'''Error thrown when trying to overwrite the obj_id'''
	def __init__(self):
		super().__init__('The attribute "_obj_id" is reserved for saving/loading and should not be set')

class SavableClassCollisionError(Exception):
	'''Error thrown when trying to re-register a class already registered'''
	def __init__(self, addr, cls):
		super().__init__('A class with the address {} is already in the class register of Packable'.format(addr))
		self.cls = cls

class UnregisteredClassError(Exception):
	'''Error thrown when trying to pack an instance of an unregistered class'''
	def __init__(self, name):
		super().__init__('"{}" is not registered (does it subclass Packable?)'.format(name))


class UnknownUserError(Exception):
	pass

class UnknownActionError(Exception):
	pass

class InsufficientPermissionsError(Exception):
	def __init__(self, user, action=None):
		msg = f'{user} is not permitted to take this action' if action is None \
			else f'{user} is not permitted to take action: {action}'
		
		super().__init__(msg)
		self.user = user
		self.action = action







from .flow import safe_self_execute, multi_index
from .logging import get_printer, get_global_setting, get_global_settings, set_global_setting, set_printer_setting
from .typing import primitives, unspecified_argument, join_classes, replace_class, \
	duplicate_class, duplicate_func, duplicate_instance, wrap_class, mix_into, \
	conditional_method, lambda_conditional_method, agnosticmethod, agnostic, agnosticproperty
from .filesystem import create_dir, crawl, spawn_path_options, load_yaml, save_yaml, \
	load_csv, load_tsv, load_json, save_json, monkey_patch, load_txt, save_txt, Persistent, HierarchyPersistent, \
	load_pickle, save_pickle
from .timing import get_now, recover_date
from .patterns import Singleton, InitWall
from .containers import deep_get, Simple_Child, Proper_Child, AttrDict, AttrOrdDict, Value, LoadedValue, Named
from .registries import Registry, Entry_Registry, Named_Registry, Class_Registry, Path_Registry, Function_Registry
from .logic import sort_by, resolve_order, toposort
from .hashing import Hashable, md5
from .utils import sign, split_dict
from .ordered_set import OrderedSet
from .farming import WorkerPool
# from .exporting import export, load_export, Exportable
from .exporting import export, load_export, set_export_manager, ExportManager, Exporter, Exportable
from .tricks import self_aware, clsdec, innerchild, method_wrapper, ClassDescriptable, classdescriptor, \
	extract_function_signature, capturable_super, captured_super, auto_init, dynamic_capture, \
	smartproperty, autoproperty, referenceproperty, defaultproperty, TrackSmart, Tracer, \
	Modifiable, inject_modifiers

from .packing import Packable, primitive, Primitive, SERIALIZABLE, JSONABLE, pack, unpack
from .packing import save_pack, load_pack, json_pack, json_unpack
# from .pure_packing import pack, unpack, json_unpack, json_pack

from .transactions import Transactionable, AbortTransaction
from .wrappers import ObjectWrapper
from .nodes import OmniNode, OmniStructure, TreeNode, LocalNode
from .viz import printc, bcolors, colorize

# from .wrappers import ObjectWrapper, Array # causes an error if required libs aren't already installed
# try:
# 	import numpy
# except ImportError:
# 	pass
# else: # Register additional common packable types
# 	from . import common

from .basic_containers import adict, tdict, tlist, tset, tstack, tdeque, theap
from .basic_containers import containerify

from .structured import TreeSpace, Table, Key_Table

import os
__info__ = {'__file__':os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py')}
with open(os.path.join(os.path.abspath(os.path.dirname(__file__)), '_info.py'), 'r') as f:
	exec(f.read(), __info__)
del os
del __info__['__file__']
__author__ = __info__['author']
__version__ = __info__['version']




from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, \
	Iterable, Iterator, TypeVar, Type
import types
import inspect
from collections import OrderedDict

from .typing import unspecified_argument


class ClassDescriptable(type):
	def __setattr__(self, key, val):
		existing = inspect.getattr_static(self, key, None)
		if isinstance(existing, classdescriptor):
			existing.__set__(self, val)
		else:
			super().__setattr__(key, val)


	def __delattr__(self, item):
		existing = inspect.getattr_static(self, item, None)
		if isinstance(existing, classdescriptor):
			existing.__delete__(self)
		else:
			super().__delattr__(item)



class classdescriptor:
	def __set__(self, obj, value):
		raise NotImplementedError


	def __delete__(self, obj):
		raise NotImplementedError



class MRONamespace(dict):
	def __init__(self, bases):
		if any(isinstance(base, MROMeta) for base in bases):
			class Scoped(*bases, use_mro_namespace=False): pass
		else:
			class Scoped(*bases): pass
		self.pseudo = Scoped

	def __getitem__(self, index):
		if index in self:
			return super().__getitem__(index)

		if hasattr(self.pseudo, index):
			return getattr(self.pseudo, index)

		raise IndexError()


class MROMeta(type):
	def __new__(metacls, name, bases, namespace, **kwargs):
		return super().__new__(metacls, name, bases, namespace)

	@classmethod
	def __prepare__(cls, name, bases, use_mro_namespace=True, **kwargs):
		if use_mro_namespace:
			return MRONamespace(bases)
		return dict()



class Scope(metaclass=MROMeta):
	pass



class method_decorator: # always replaces the function
	def __init__(self, fn: Callable = None, *, enforce_setup: bool = True):
		self.fn = fn
		self._enforce_setup = enforce_setup
		self._is_setup = False

	def __call__(self, fn: Callable) -> 'method_decorator':
		self.fn = fn
		return self


	def setup(self, owner: Type, name: Optional[str] = None) -> 'method_decorator':
		self._is_setup = True
		if name is None:
			name = self.fn.__name__
		self._setup(owner, name)
		return self


	def _setup(self, owner: Type, name: str) -> None:
		pass


	def __set_name__(self, owner: Type, name: str) -> None:
		setattr(owner, name, self.setup(owner, name))


	def __get__(self, instance: Any, owner: Type) -> Any:
		if self._enforce_setup and not self._is_setup:
			raise RuntimeError(f'{self.__class__.__name__} not setup properly '
							   f'(call .setup(name, owner) or use as decorator)')
		return self.package(self.fn, instance, owner)


	@staticmethod
	def package(fn: Callable, instance: Any, owner: Type) -> Callable:
		getter = getattr(fn, '__get__', None)
		if getter is not None and instance is not None:
			return getter(instance, owner)
		return fn


class nested_method_decorator(method_decorator):
	def _setup(self, owner: Type, name: str) -> None:
		set_name = getattr(self.fn, '__set_name__', None)
		if set_name is not None:
			set_name(owner, name)
		return super()._setup(owner, name)

	def __get__(self, instance: Any, owner: Type) -> Any:
		if self._enforce_setup and not self._is_setup:
			raise RuntimeError(f'{self.__class__.__name__} not setup properly '
							   f'(call .setup(name, owner) or use as decorator)')
		fn = self.fn
		getter = getattr(fn, '__get__', None)
		if getter is not None:
			fn = getter(instance, owner)
		return self.package(fn, instance, owner)


class method_binder(nested_method_decorator):
	class future_method:
		def __init__(self, src: Type, fn: Callable, owner: Type, instance: Any = None, *,
		             is_static: Optional[bool] = None, **kwargs):
			'''
			
			Args:
				src: Class where the method is defined
				fn: Method to call
				owner: Class where the method is called
				instance: Instance where the method is called
				is_static: Whether the method is a staticmethod or classmethod
			'''
			# if is_static is None:
			# 	is_static = fn in {staticmethod, classmethod}
			self.fn = fn
			self.instance = instance
			self.src = src
			self.owner = owner
			# self.is_static = is_static
			

		def __repr__(self):
			return f'<future_method {self.fn.__name__} of {self.owner.__name__}>' if self.instance is None \
				else f'<future_method {self.fn.__name__} of {self.owner.__name__} bound to {self.instance}>'

		def __call__(self, *args: Any, **kwargs: Any) -> Any:
			# if not self.is_static and self.instance is None:
			# 	assert len(args), 'no instance to call method on'
			# 	self.instance = args[0]
			# 	args = args[1:]
			return self.fn_call(self.fn,
			                    # None if self.is_static else
			                    self.instance,
			                    *args, **kwargs)

		@staticmethod
		def fn_call(fn: Callable, instance: Any, *args, **kwargs) -> Any:
			return fn(*args, **kwargs)
	
	def _setup(self, owner: Type, name: str) -> None:
		self.src = owner
		return super()._setup(owner, name)
	
	def package(self, fn: Callable, instance: Any, owner: Type) -> future_method:
		return self.future_method(self.src, fn, owner, instance,)
		                          # is_static=isinstance(self.fn, (staticmethod, classmethod)))



class capturable_method:
	@classmethod
	def captured_method_call(cls, self: Optional['capturable_method'], src: Type, fn: Callable,
	                         args: Tuple, kwargs: Dict[str, Any]) -> Any:
		return fn(*args, **kwargs)


class captured_method(method_binder):
	_capturer_type = None

	class future_method(method_binder.future_method):
		def fn_call(self, fn: Callable, instance: Any, *args: Any, **kwargs: Any) -> Any:
			return self.owner.captured_method_call(instance, self.src, fn, args, kwargs)


	# def package(self, instance: Any, owner: Type) -> future_method:
	# 	return self.future_method(self.fn, self.location, instance)




class _capturable_super_parent: pass # hidden parent class to hold the delegators
class capturable_super(_capturable_super_parent):
	@classmethod
	def captured_super_call(cls, self, src: Type, name: str, args: Tuple, kwargs: Dict[str, Any]) -> Any:
		'''
		Called when methods decorated with @capture_super call super() (without parameters).

		:param src: the class from which the method was called
		:param name: of the method being called
		:param args: positional arguments passed to the method
		:param kwargs: keyword arguments passed to the method
		:return: output of the desired effect of super().[name](*args, **kwargs)
		'''
		return getattr(super(src, self), name)(*args, **kwargs)



class captured_super(method_binder):
	_child_capturer = capturable_super # gets set as the __class__ of methods decoared with @capture_super
	_parent_capturer = _capturable_super_parent # contains the corresponding delegator to execute capture


	def _setup(self, owner: Type, name: str) -> None:
		self.name = name
		if self.fn is not None:
			setattr(self._parent_capturer, self.name, self.run_delegator(self.name))

		# if self.fn is not None and self.fn.__closure__ is not None:
		# 	self.fn.__closure__[0].cell_contents = self._child_capturer
		return super()._setup(owner, name)


	class future_method(method_binder.future_method):
		def __init__(self, src, fn, owner, instance=None, flag=None):
			super().__init__(src, fn, owner, instance)
			self.flag = flag

		def fn_call(self, fn, instance, *args, **kwargs):
			if instance is None:
				raise NotImplementedError('capturing super() for classmethods or staticmethods '
				                          'is currently not supported.')
				
			setattr(instance, self.flag, self.owner)
			return super().fn_call(fn, instance, *args, **kwargs)


	def package(self, fn: Callable, instance: Any, owner: Type) -> future_method:
		if fn is not None and fn.__closure__ is not None:
			fn.__closure__[0].cell_contents = self._child_capturer
		return self.future_method(self.src, fn, owner, instance,
			self.run_delegator.single_run.instance_trigger.format(name=self.name))


	class run_delegator: # located in _parent_capturer and checks if instance can trigger a capture of a method
		def __init__(self, name: str):
			self.name = name

		class single_run: # created by delegator for a single run with a specific instance and method
			instance_trigger = '_invisible_super_trigger_for_{name}'

			def __init__(self, obj, cls, name):
				self.obj = obj
				self.cls = cls
				self.name = name

			def clean_up(self):
				key = self.instance_trigger.format(name=self.name)
				if hasattr(self, key):
					delattr(self.obj, key)

			def __call__(self, *args, **kwargs):
				out = self.run(self.cls, self.obj, self.name, args, kwargs)
				self.clean_up()
				return out

			@staticmethod
			def run(cls, obj, name, args, kwargs):
				return obj.captured_super_call(cls, name, args, kwargs)

		def __get__(self, instance: Any, owner: Type) -> Union[single_run, Callable]:
			if instance is None:
				raise AttributeError(self.name)

			true_owner = getattr(instance, self.single_run.instance_trigger.format(name=self.name), None)
			if true_owner is None:
				# return getattr(instance, self.name)
				return getattr(super(owner, instance), self.name)
				# raise AttributeError(self.name)

			return self.single_run(instance, true_owner, self.name)




class Capturable(capturable_method, capturable_super):
	pass



class captured(captured_super, captured_method):
	class future_method(captured_super.future_method, captured_method.future_method): pass



class method_wrapper(nested_method_decorator):
	def package(self, fn: Callable, instance: Any, owner: Type = None) -> Callable:
		return self.method_application(self, fn, instance, owner)
	
	@staticmethod
	def process_args(args, kwargs, owner, instance, fn):
		return args, kwargs

	@staticmethod
	def process_out(out: Any, owner, instance, fn) -> Any:
		return out

	class method_application:
		def __init__(self, wrapper: 'method_wrapper', fn: Callable, instance: Any, owner: Type = None):
			self.wrapper, self.fn, self.instance, self.owner = wrapper, fn, instance, owner

		def __call__(self, *args, **kwargs):
			args, kwargs = self.wrapper.process_args(args, kwargs, self.owner, self.instance, self.fn)
			out = self.fn(*args, **kwargs)
			return self.wrapper.process_out(out, self.owner, self.instance, self.fn)



class auto_methods(capturable_method):
	# _auto_methods = None
	def __init_subclass__(cls, auto_methods: Optional[Union[str, Sequence[str]]] = (), wrap_existing: bool = False,
	                      inheritable_auto_methods: Optional[Union[str, Sequence[str]]] = (), **kwargs):
		super().__init_subclass__(**kwargs)

		existing = [method for method, inherit in getattr(cls, '_auto_methods', {}).items() if inherit]
		if isinstance(inheritable_auto_methods, str):
			inheritable_auto_methods = (inheritable_auto_methods,)
		if isinstance(auto_methods, str):
			auto_methods = (auto_methods,)
		cls._auto_methods = OrderedDict((method, True) for method in existing + list(inheritable_auto_methods))
		cls._auto_methods.update((method, False) for method in auto_methods)

		for method in cls._auto_methods:
			if wrap_existing or method in cls.__dict__:
				setattr(cls, method, captured_method(inspect.getattr_static(cls, method)).setup(cls, method))


	@classmethod
	def _auto_method_call(cls, self, src: Type, method: Callable, args: Tuple, kwargs: Dict[str, Any]) -> None:
		raise NotImplementedError

	@classmethod
	def captured_method_call(cls, self, src: Type['auto_methods'], fn: Callable,
	                         args: Tuple, kwargs: Dict[str, Any]) -> Any:
		if getattr(src, '_auto_methods', None) is not None and fn.__name__ in src._auto_methods:
			return cls._auto_method_call(self, src, fn, args, kwargs)
		return super().captured_method_call(self, src, fn, args, kwargs)


class auto_fix_args(auto_methods):
	class _auto_method_arg_fixer:
		def __init__(self, method: Callable, src: Type, owner: Type, obj: 'auto_methods', **kwargs):
			super().__init__(**kwargs)
			self.method = method
			self.src = src
			self.owner = owner
			self.obj = obj

		def __call__(self, key: str, default: Optional[Any] = inspect.Parameter.empty) -> Any:
			raise KeyError(key)

	class MissingArgumentsError(TypeError):
		def __init__(self, src, method, missing, *, msg=None):
			if msg is None:
				msg = f'{src.__name__}.{method.__name__}() missing {len(missing)} ' \
				      f'required arguments: {", ".join(missing)}'
			super().__init__(msg)
			self.missing = missing
			self.src = src
			self.method = method

	@classmethod
	def _auto_method_call(cls, self, src: Type, method: Callable, args: Tuple, kwargs: Dict[str, Any]) -> None:
		fixed_args, fixed_kwargs, missing = extract_function_signature(method, args, kwargs,
												default_fn=cls._auto_method_arg_fixer(method, src, cls, self),
											                  include_missing=True)
		if len(missing):
			raise cls.MissingArgumentsError(src, method, missing)
		return method(*fixed_args, **fixed_kwargs)


class auto_init(auto_methods, inheritable_auto_methods='__init__'):
	pass


T = TypeVar('T')

class dynamic_capture:
	def __init__(self, srcs: Sequence[Type[T]],
	             fn: Callable[[Type, Callable, T, Tuple, Dict[str, Any]], Any],
	             *method_names: str):

		self.srcs = srcs
		self.fn = fn
		self.method_names = method_names

		self.original_methods = None

		# if not isinstance(owner, type):
		# 	owner = type(owner)


	class capture_context:
		def __init__(self, owner, method_fn, fn):
			self.owner = owner # current class
			self.instance = None # instance of owner
			self.method_fn = method_fn # original method of src that was now replaced with self
			self.fn = fn

		def __get__(self, instance, owner):
			self.instance = instance
			return self
			# return self.fn(instance, self.src, self.original)

		def __call__(self, *args, **kwargs):
			# assert self.instance is not None, 'cannot call a captured method without an instance'
			return self.fn(self.owner, self.method_fn, self.instance, args, kwargs)


	def activate(self):
		original_methods = {}
		for cls in self.srcs:
			if cls is not object:
				for name in self.method_names:
					if name in cls.__dict__:
						original_methods[cls, name] = getattr(cls, name)
						setattr(cls, name, self.capture_context(cls, original_methods[cls, name], self.fn))
		self.original_methods = original_methods
		return self

	class NotActivatedError(ValueError): pass

	def deactivate(self):
		if self.original_methods is None:
			raise self.NotActivatedError('cannot deactivate a dynamic capture that has not been activated')
		for (cls, name), original in self.original_methods.items():
			setattr(cls, name, original)

	def __enter__(self):
		return self.activate()

	def __exit__(self, exc_type, exc_val, exc_tb):
		self.deactivate()
		# return exc_type, exc_val, exc_tb


class simple_dynamic_capture(dynamic_capture):
	def __init__(self, target: Type[T], fn: Callable[[Type, Callable, T, Tuple, Dict[str, Any]], Any],
	             *method_names: str, full_bases: Optional[bool] = False, full_mro: Optional[bool] = False):
		super().__init__(target.mro() if full_mro else target.__bases__ if full_bases else (target,),
		                 fn, *method_names)




class smartproperty:
	unknown = object()

	def __init__(self, *, name=None, src=None, cache=False, fget=None, fset=None, fdel=None, doc=None):
		# super().__init__(fget=fget, **kwargs)
		self.name = name
		self.src = src

		self.cache = cache
		self.cached_value = self.unknown

		self.fget = fget
		self.fset = fset
		self.fdel = fdel
		self.doc = doc

		self._during_getter = False
		self._during_setter = False
		self._during_deleter = False

	def copy(self, src=unspecified_argument, name=unspecified_argument,
	         default=unspecified_argument, cache=unspecified_argument,
	         fget=unspecified_argument, fset=unspecified_argument, fdel=unspecified_argument, doc=unspecified_argument,
	         **kwargs):
		if src is unspecified_argument:
			src = self.src
		if name is unspecified_argument:
			name = self.name
		if default is unspecified_argument:
			default = self.default
		if cache is unspecified_argument:
			cache = self.cache
		if fget is unspecified_argument:
			fget = self.fget
		if fset is unspecified_argument:
			fset = self.fset
		if fdel is unspecified_argument:
			fdel = self.fdel
		if doc is unspecified_argument:
			doc = self.__doc__
		return type(self)(name=name, src=src, default=default, cache=cache,
		                  fget=fget, fset=fset, fdel=fdel, doc=doc, **kwargs)

	def __set_name__(self, owner, name):
		if self.src is not None and owner is not self.src:
			setattr(owner, name, self.copy(src=owner, name=name))
		else:
			if self.name is None: # TODO: test with using subsequent .setter as with properties
				self.name = name
			self.src = owner
			setattr(owner, name, self)

	def __call__(self, fget: Callable[[], Any]) -> 'smartproperty':
		return self.getter(fget)

	def getter(self, fget: Callable[[], Any]) -> 'smartproperty':
		self.fget = fget
		return self

	def setter(self, fset: Callable[[Any], None]) -> 'smartproperty':
		self.fset = fset
		return self

	def deleter(self, fdel: Callable[[], None]) -> 'smartproperty':
		self.fdel = fdel
		return self

	def __get__(self, instance, owner=None):
		return self.get_value(instance, owner)

	def __set__(self, instance, value) -> None:
		self.update_value(instance, value)

	def __delete__(self, instance):  # TODO: test this
		self.reset(instance)

	class MissingValueError(AttributeError):
		def __init__(self, base, name, *, msg=None):
			basename = base.__name__ if isinstance(base, type) else base.__class__.__name__
			super().__init__(f'{basename}.{name} has no value' if msg is None else msg)

	class GetterFailedError(MissingValueError):
		def __init__(self, base, name=None, *, msg=None):
			basename = base.__name__ if isinstance(base, type) else base.__class__.__name__
			super().__init__(f'{basename}{"" if name is None else "."+name}' if msg is None else msg)

	# class SetterFailedError(TypeError):
	# 	def __init__(self, base, name=None, value=None, *, msg=None):
	# 		basename = base.__name__ if isinstance(base, type) else base.__class__.__name__
	# 		super().__init__(f'{basename}{"" if name is None else "."+name} setter failed' if msg is None else msg)
	#
	# class DeleterFailedError(TypeError):
	# 	def __init__(self, base, name=None, *, msg=None):
	# 		basename = base.__name__ if isinstance(base, type) else base.__class__.__name__
	# 		super().__init__(f'{basename}{"" if name is None else "."+name} deleter failed' if msg is None else msg)

	@staticmethod
	def _call_descriptor(fn, instance, owner, *args, **kwargs):
		getter = getattr(fn, '__get__', None)
		if getter is not None:
			fn = getter(instance, owner)
		return fn(*args, **kwargs)

	def create_value(self, base, owner=None):
		if self.fget is None:
			raise self.MissingValueError(base, self.name)
		try:
			return self._call_descriptor(self.fget, base, owner)
		except self.GetterFailedError:
			raise self.MissingValueError(base, self.name) from None

	def _set_cached_value(self, base, value):
		if base is self.src:
			self.cached_value = value
			return
		elif isinstance(base, type):
			raise NotImplementedError

		setter = getattr(base, '__setattr__', None)
		if setter is not None and not self._during_setter:
			self._during_setter = True
			setter(self.name, value)
			self._during_setter = False
		else:
			# raise AttributeError(f'cannot cache attribute {self.name} of {base}')
			# raise NotImplementedError
			cache = getattr(base, '__dict__', None)
			if cache is None or self.name is None:
				raise AttributeError(f'cannot cache attribute {self.name} of {base}')
			if cache is not None:
				cache[self.name] = value

	def _get_cached_value(self, base):
		if base is self.src:
			return self.cached_value
		elif isinstance(base, type):
			raise NotImplementedError

		getter = getattr(base, '__getattr__', None)
		if getter is not None and not self._during_getter:
			self._during_getter = True
			try:
				value = getter(self.name)
			except AttributeError:
				pass
			else:
				return value
			finally:
				self._during_getter = False

		cache = getattr(base, '__dict__', None)
		if cache is None:
			return self.unknown
		return cache.get(self.name, self.unknown)

	def _clear_cache(self, base):
		if base is self.src:
			self.cached_value = self.unknown
			return
		elif isinstance(base, type):
			raise NotImplementedError

		deleter = getattr(base, '__delattr__', None)
		if deleter is not None and not self._during_deleter:
			self._during_deleter = True
			deleter(self.name)
			self._during_deleter = False
			return
		else:
			# raise AttributeError(f'cannot reset attribute {self.name} of {base}')
			# raise NotImplementedError
			cache = getattr(base, '__dict__', None)
			if cache is None or self.name is None:
				raise AttributeError(f'cannot reset attribute {self.name} of {base}')
			if cache is not None:
				cache.pop(self.name, None)

	def get_value(self, base, owner=None): # TODO: maybe make thread-safe by using a lock
		'''
		Check cache of base first, if that fails, then try getter function

		Args:
			base: instance or type to get value from
			owner: type of instance if it is provided

		Returns:
			value of property

		'''
		value = self._get_cached_value(base)
		if value is self.unknown:
			value = self.create_value(base, owner)
			if self.cache:
				self._set_cached_value(base, value)
		return value
		
	def update_value(self, base, value):
		'''Try manually specified setter function, if that fails, then try to set value in cache'''
		if self.fset is None:
			self._set_cached_value(base, value)
		else:
			return self._call_descriptor(self.fset, base, None, value)

	def reset(self, base):
		if self.fdel is None:
			self._clear_cache(base)
		else:
			return self._call_descriptor(self.fdel, base, None)



# class cachedproperty(smartproperty):
# 	def __init__(self, fget: Callable[[], Any] = None, *, cache=False, **kwargs):
# 		super().__init__(fget=fget, **kwargs)
# 		self.cache = cache
# 		self.cached_value = self.unknown
#
# 	def _get_cached_value(self, base):
# 		if base is self.src:
# 			return self.cached_value
# 		return super()._get_cached_value(base)
#
# 	def _set_cached_value(self, base, value):
# 		if base is self.src:
# 			self.cached_value = value
# 		else:
# 			super()._set_cached_value(base, value)
#
# 	def _clear_cache(self, base):
# 		if base is self.src:
# 			self.cached_value = self.unknown
# 		else:
# 			super()._clear_cache(base)
#
# 	def get_value(self, base, owner=None):
# 		value = self._get_cached_value(base)
# 		if value is self.unknown:
# 			value = self.create_value(base, owner)
# 			if self.cache:
# 				self._set_cached_value(base, value)
# 		return value


class autoproperty(smartproperty): # agnostic to whether it is a class or instance attribute
	def _get_base(self, instance, owner=None):
		return owner if instance is None else instance

	def __get__(self, instance, owner=None):
		return self.get_value(self._get_base(instance, owner), owner)

	def __set__(self, instance, value) -> None:
		self.update_value(self._get_base(instance), value)

	def __delete__(self, instance):  # TODO: test this
		self.reset(self._get_base(instance))
	
	
class _referenceproperty_target:
	def __init__(self, prop, base):
		self.prop = prop
		self.base = base

	def _get_reference(self):
		ref = getattr(self.base, self.prop.ref_name, None)
		# if ref is None:
		# 	raise self.prop.MissingReferenceError(f'{self.base} has no reference to {self.prop.ref_name}')
		return ref

	@staticmethod
	def target_call(ref, *args, **kwargs):
		raise NotImplementedError

	def __call__(self, *args, **kwargs):
		return self.target_call(self._get_reference(), *args, **kwargs)


class referenceproperty(smartproperty):

	class MissingReferenceError(AttributeError):
		pass
	
	class _reference_getter(_referenceproperty_target):
		def target_call(self, ref):
			if ref is None:
				raise self.prop.GetterFailedError(self.base, self.prop.name)
			return getattr(ref, self.prop.name)
	
	class _reference_setter(_referenceproperty_target):
		def target_call(self, ref, value):
			setattr(ref, self.prop.name, value)
	
	class _reference_deleter(_referenceproperty_target):
		def target_call(self, ref):
			delattr(ref, self.prop.name)
	
	class _reference_descriptor:
		def __init__(self, prop, target):
			self.prop = prop
			self.target = target

		def __get__(self, instance, owner=None):
			return self.target(self.prop, instance)

	def __init__(self, ref_key: str, *, use_getter=True, use_setter=False, use_deleter=False, **kwargs):
		super().__init__(**kwargs)
		self.ref_key = ref_key
		self.use_getter = use_getter
		self.use_setter = use_setter
		self.use_deleter = use_deleter
		
	@property
	def fget(self):
		if self._fget is None and self.use_getter:
			return self._reference_descriptor(self, self._reference_getter)
		return self._fget
	@fget.setter
	def fget(self, value):
		self._fget = value
	
	@property
	def fset(self):
		if self._fset is None and self.use_setter:
			return self._reference_descriptor(self, self._reference_setter)
		return self._fset
	@fset.setter
	def fset(self, value):
		self._fset = value
	
	@property
	def fdel(self):
		if self._fdel is None and self.use_deleter:
			return self._reference_descriptor(self, self._reference_deleter)
		return self._fdel
	@fdel.setter
	def fdel(self, value):
		self._fdel = value
	


class defaultproperty(smartproperty):
	def __init__(self, default=unspecified_argument, *, fget=unspecified_argument, **kwargs):
		if default is unspecified_argument:
			default = self.unknown
		fget, default = self._check_fget(fget, default)
		super().__init__(fget=fget, **kwargs)
		self.default = default

	def _check_fget(self, fget=unspecified_argument, default=unspecified_argument):
		if fget is unspecified_argument:
			if callable(default) and not isinstance(default, type) and default.__qualname__ != default.__name__:
				return default, unspecified_argument  # decorator has no other args
			return None, default  # no fget provided (optionally can be added with __call__)
		return fget, default  # fget was specified as keyword argument

	def create_value(self, base, owner=None): # TODO: maybe make thread-safe by using a lock
		try:
			return super().create_value(base, owner)
		except self.MissingValueError:
			if self.default is self.unknown:
				raise
			return self.default


from .utils import split_dict


class TrackSmart:
	def _extract_smart_properties(self, kwargs, *, src=None):
		if src is None:
			src = dict(self._my_smart_properties())
		found, remaining = split_dict(kwargs, src)
		for key, val in found.items():
			setattr(self, key, val)
		return remaining

	def _get_property(self, name, default=unspecified_argument):
		try:
			return inspect.getattr_static(self, name)
		except AttributeError:
			if default is unspecified_argument:
				raise
			return default

	# def _fill_smart_properties(self, kwargs, *, src=None):
	# 	if src is None:
	# 		src = dict(self._my_smart_properties())
	# 	for key, val in src.items():
	# 		if key not in kwargs:
	# 			kwargs[key] = val

	@classmethod
	def _my_smart_properties(cls):
		for name, value in cls.__dict__.items():
			if isinstance(value, smartproperty):
				yield name, value

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **self._extract_smart_properties(kwargs))


class Tracer(tuple):
	def append(self, term):
		return self.__class__((self, term))

	def extend(self, terms):
		for term in terms:
			self = self.append(term)
		return self

	def back(self, n=1, strict=True):
		if n == 0:
			return self
		if n < 0:
			raise ValueError('n must be >= 0')
		if not self:
			if strict:
				raise IndexError('cannot go back from empty tracer')
			return self
		return self[0].back(n - 1)

	@property
	def path(self):
		if len(self):
			return self[0].path + (self[1],)
		return ()



def inject_modifiers(base, *mods, name=None):
	'''
	Mods should be types that are used to modify the cls. The order of the mods corresponds to the order
	in which the cls is "modified", so for example:

	```python

	class A(Modifiable): pass
	class B: pass
	class C: pass

	out = A.inject_mods(B, C)
	assert out.mro() == [out, C, B, A, Modifiable, object]
	assert out.__name__ == 'C_B_A' # default name
	```
	'''
	if len(mods):
		bases = (*reversed(mods), base)
		if name is None:
			name = '_'.join(base.__name__ for base in bases)
		return type(name, bases, {})
	return base


class Modifiable:
	@classmethod
	def inject_mods(cls, *mods, name=None):
		'''
		Mods should be types that are used to modify the cls. The order of the mods corresponds to the order
		in which the cls is "modified", so for example:

		```python

		class A(Modifiable): pass
		class B: pass
		class C: pass

		out = A.inject_mods(B, C)
		assert out.mro() == [out, C, B, A, Modifiable, object]
		assert out.__name__ == 'C_B_A' # default name
		```
		'''
		return inject_modifiers(cls, *mods, name=name)


# class old_auto_init(capturable_method):
# 	def __init_subclass__(cls, **kwargs):
# 		super().__init_subclass__(**kwargs)
# 		if '__init__' in cls.__dict__:
# 			cls.__init__ = captured_method(cls.__init__).setup(cls)
#
#
# 	class _init_arg_fixer:
# 		def __init__(self, src: Type, obj: 'old_auto_init', **kwargs):
# 			super().__init__(**kwargs)
# 			self.src = src
# 			self.obj = obj
#
# 		def __call__(self, key: str, default: Optional[Any] = inspect.Parameter.empty) -> Any:
# 			raise KeyError(key)
#
#
# 	def _auto_init(self, src: Type, init_fn: Callable, args: Tuple, kwargs: Dict[str, Any]) -> None:
# 		fixed_args, fixed_kwargs = extract_function_signature(init_fn, (self, *args), kwargs,
# 															  default_fn=self._init_arg_fixer(src, self))
# 		return init_fn(*fixed_args, **fixed_kwargs)
#
#
# 	def captured_method_call(self, src: Type, fn: Callable, args: Tuple, kwargs: Dict[str, Any]) -> Any:
# 		if fn.__name__ == '__init__':
# 			return self._auto_init(src, fn, args, kwargs)
# 		return super().captured_method_call(src, fn, args, kwargs)




# Example usage:

# class A(Capturable):
#     def captured_super_call(self, src, name, args, kwargs):
#         print('!! captured super', self, src, name, args, kwargs)
#         return super().captured_super_call(src, name, args, kwargs)
#
#     def captured_method_call(self, src, fn, args, kwargs):
#         print('!! captured method', self, src, fn, args, kwargs)
#         return super().captured_method_call(src, fn, args, kwargs)
#
#     def f(self, a=1):
#         print('A.f', self, a)
#         # print(super().f())
#
# class B(A):
#     x = -10
#
#     def f(self, a=2):
#         print('B.f', self, a)
#         super().f()
#
# class C(B):
#     @captured
#     def f(self, a=3):
#         print('C.f', self, a)
#         super().f()
#
# class D(C):
#     def f(self, a=4):
#         print('D.f', self, a)
#         super().f()
#
# class E(D):
#     @captured
#     def f(self, a=5):
#         print('E.f', self, a)
#         super().f()
#
# class F(E):
#     def f(self, a=6):
#         print('F.f', self, a)
#         super().f()
#
# class G(F):
#     pass
#
# A().f()
# print()
# B().f()
# print()
# C().f()
# print()
# D().f()
# print()
# E().f()
# print()
# F().f()
# print()
# G().f()

###################################################




class self_aware:
	def __init__(self, cls: Type) -> None:
		self.cls = cls

	def __set_name__(self, owner: Type, name: str) -> None:
		self.cls.owner = owner
		setattr(owner, name, self.cls)



class _meta_clsdec(type):
	def __getattr__(self, key: str) -> '_meta_clsdec':
		self.key = key
		return self



class clsdec(metaclass=_meta_clsdec):
	def __init__(self, *args: Any, **kwargs: Any) -> None:
		self.args, self.kwargs = args, kwargs


	def __call__(self, fn: Callable) -> 'clsdec':
		self.fn = fn
		return self


	def __set_name__(self, obj: Any, name: str) -> None:
		setattr(obj, name, getattr(obj, self.key)(*self.args, **self.kwargs)(self.fn))



class innerchild:
	def __init__(self, cls: Type) -> None:
		self.cls = cls


	class MissingParent(Exception):
		def __init__(self, base: Type, name: str) -> None:
			super().__init__(f'{base.__name__} has not inner class {name}')


	def __set_name__(self, owner: Type, name: str) -> None:
		cls_name = self.cls.__name__
		parent = getattr(super(owner, owner), cls_name, None)
		if parent is None:
			raise self.MissingParent(owner, cls_name)
		child = type(cls_name, (self.cls, parent), {})
		setattr(owner, name, child)


from inspect import Parameter

def deep_method_finder(typ: Type, method: str, *, break_fn: Callable[[Type],bool] = None) -> Iterator[Callable]:
	on_parents = False
	for cls in typ.mro():
		fn = cls.__dict__.get(method, None)
		if fn is not None:
			if on_parents and break_fn is not None and break_fn(fn):
				break
			on_parents = True
			yield fn

def collect_fn_kwargs(*fns: Callable, default: Any = Parameter.empty, ignore_positional_only=False):
	kwargs = {}

	for fn in fns:
		params = inspect.signature(fn).parameters
		for param in params.values():
			if param.kind == param.POSITIONAL_ONLY and not ignore_positional_only:
				raise TypeError(f'Positional only arguments are not supported: {param.name}')
			elif param.kind in {param.POSITIONAL_OR_KEYWORD, param.KEYWORD_ONLY} and param.name not in kwargs:
				kwargs[param.name] = default if param.default is param.empty else param.default

	return kwargs

def collect_init_kwargs(typ: Type, default: Any = Parameter.empty, *, end_type: Union[Sequence[Type], Type] = None,
						ignore_positional_only=False, break_fn=None):
	if break_fn is None:
		if end_type is not None and isinstance(end_type, type):
			end_type = {end_type,}
		break_fn = (lambda cls: True) if end_type is None else (lambda cls: cls in end_type)
	return collect_fn_kwargs(*deep_method_finder(typ, method='__init__', break_fn=break_fn),
							 default=default, ignore_positional_only=ignore_positional_only)

	
def extract_function_signature(fn: Union[Callable, Type],
                               args: Optional[Tuple] = None, kwargs: Optional[Dict[str, Any]] = None, *,
                               default_fn: Callable[[str, Any], Any] = None, include_missing: bool = False,
                               allow_positional: bool = True) \
		-> Union[Tuple[List[Any], Dict[str, Any], List[inspect.Parameter]],
		         Tuple[List[Any], Dict[str, Any]],
		         Dict[str, Any]]:
	if args is None:
		args = ()
	if kwargs is None:
		kwargs = {}

	if isinstance(fn, type):
		fn = fn.__init__

	params = inspect.signature(fn).parameters
	
	arg_idx = 0
	fixed_args = []
	fixed_kwargs = {}

	missing = []

	for n, p in params.items():
		if p.kind == p.POSITIONAL_ONLY:
			if not allow_positional:
				raise TypeError(f'Function {fn.__name__} has a positional only argument ({n})')
			if arg_idx < len(args):
				fixed_args.append(args[arg_idx])
				arg_idx += 1
			else:
				try:
					if default_fn is None:
						raise KeyError
					val = default_fn(n, p.default)
				except KeyError:
					if p.default is p.empty:
						missing.append(p)
				else:
					fixed_args.append(val)
		elif p.kind == p.VAR_POSITIONAL:
			if not allow_positional:
				# raise TypeError(f'Function {fn.__name__} has variable positional arguments ({n})')
				continue
			try:
				if default_fn is None:
					raise KeyError
				val = default_fn(n, p.default)
			except KeyError:
				fixed_args.extend(args[arg_idx:])
				arg_idx = len(args)
			else:
				fixed_args.extend(val)
		elif p.kind == p.VAR_KEYWORD:
			try:
				if default_fn is None:
					raise KeyError
				val = default_fn(n, p.default)
			except KeyError:
				fixed_kwargs.update(kwargs)
			else:
				fixed_kwargs.update(val)

		else:
			if n in kwargs:
				fixed_kwargs[n] = kwargs[n]
			elif p.kind != p.KEYWORD_ONLY and arg_idx < len(args):
				if allow_positional:
					fixed_args.append(args[arg_idx])
				else:
					fixed_kwargs[n] = args[arg_idx]
				arg_idx += 1
			else:
				try:
					if default_fn is None:
						raise KeyError
					val = default_fn(n, p.default)
				except KeyError:
					if p.default is p.empty:
						missing.append(p)
					# if p.default is p.empty:
					# 	raise TypeError(f'Argument {n} is missing')
					# print(n, p.default)
					pass
				else:
					fixed_kwargs[n] = val
	if include_missing:
		return fixed_args, fixed_kwargs, missing
	if allow_positional:
		return fixed_args, fixed_kwargs
	assert len(fixed_args) == 0, f'fixed_args: {fixed_args}'
	return fixed_kwargs



# class Property(object):
#     "Emulate PyProperty_Type() in Objects/descrobject.c"
#
#     def __init__(self, fget=None, fset=None, fdel=None, doc=None):
#         self.fget = fget
#         self.fset = fset
#         self.fdel = fdel
#         if doc is None and fget is not None:
#             doc = fget.__doc__
#         self.__doc__ = doc
#
#     def __get__(self, obj, objtype=None):
#         if obj is None:
#             return self
#         if self.fget is None:
#             raise AttributeError("unreadable attribute")
#         return self.fget(obj)
#
#     def __set__(self, obj, value):
#         if self.fset is None:
#             raise AttributeError("can't set attribute")
#         self.fset(obj, value)
#
#     def __delete__(self, obj):
#         if self.fdel is None:
#             raise AttributeError("can't delete attribute")
#         self.fdel(obj)
#
#     def getter(self, fget):
#         return type(self)(fget, self.fset, self.fdel, self.__doc__)
#
#     def setter(self, fset):
#         return type(self)(self.fget, fset, self.fdel, self.__doc__)
#
#     def deleter(self, fdel):
#         return type(self)(self.fget, self.fset, fdel, self.__doc__)





# class Grandparent:
# 	print('setting up Grandparent')
# 	class mydec:
# 		def __init__(self, *args, **kwargs):
# 			print('Grandparent.mydec.__init__')
#
# 		def __call__(self, fn):
# 			print('Grandparent.mydec.__call__')
# 			return fn
#
#
# class Uncle(Grandparent):
# 	print('setting up Uncle')
# 	@clsdec.mydec('dec_arg_uncle', dec_kwarg=2)
# 	def do_something(self, x=1, y=2):
# 		print('Uncle.do_something')
#
#
# class Parent(Grandparent):
# 	print('setting up Parent')
# 	@innerchild
# 	class mydec:
# 		def __init__(self, *args, **kwargs):
# 			print('Parent.mydec.__init__')
# 		def __call__(self, fn):
# 			print('Parent.mydec.__call__')
# 			return super().__call__(fn)
#
#
# class Child(Parent):
# 	print('setting up Child')
# 	@clsdec.mydec('dec_arg_child', dec_kwarg=1)
# 	def do_something(self, x=3, y=4):
# 		print('Child.do_something')








import traceback

from .basic_containers import adict, tlist




class TreeSpace(adict):
	'''
	Namespace - like a dictionary but where keys can be accessed as attributes, and if not found will create new NS
	allowing:

	a = TreeSpace()
	a.b.c.d = 'hello'
	print(repr(a)) # --> {{'b':{{'c':{{'d':'hello'}}}}}}

	NOTE: avoid ``hasattr``! - always returns true (creating new attrs), use ``__contains__`` instead

	'''
	
	def __getitem__(self, key):
		try:
			v = super().__getitem__(key)
			# print(key,v)
			return v
		except KeyError:
			try:
				return super().__getattribute__(key)
			except AttributeError:
				# print('**WARNING: defaulting {}'.format(key))
				return self._missing_key(key)
	
	def _missing_key(self, key):
		obj = self.__class__()
		self.__setitem__(key, obj)
		return obj
	
	def todict(self):
		d = {}
		for k ,v in self.items():
			if isinstance(v, TreeSpace):
				v = v.todict()
			d[k] = v
		return d
	
	def __repr__(self):
		return '{}{}{}'.format('{{', ', '.join(['{}:{}'.format(repr(k), repr(v)) for k ,v in self.items()]), '}}')

class Table(tlist):
	'''
	Essentially a database (elements are rows, keys are cols)
	Allowing nonrectangular entries
	All elements should be dicts (or ideally tdicts)
	'''
	
	def __init__(self, *args, _type=adict, **kwargs):
		super().__init__(*args, **kwargs)
		self.__dict__['_el_type'] = _type
	
	def sort_by(self, *keys):
		self.__dict__['_data'] = sorted(self, key=lambda x: tuple(x[k] for k in keys))
	
	def select(self, key, skip=True):
		for x in self.selects(key, skip=skip):
			yield x[0]
	
	def selects(self, *keys, skip=True):
		for x in self:
			l = []
			for k in keys:
				if k in x:
					l.append(x[k])
				elif skip:
					l = None
					break
				else:
					raise KeyError(k)
			if l is not None:
				yield l
	
	def select_items(self, key, skip=True, skip_flag=None):
		for x in self:
			if key in x:
				yield x[key], x
			elif not skip:
				yield skip_flag, x
	
	def _join(self, other, cmp, merge_fn=None):
		
		if isinstance(cmp, str):
			cmp = lambda a ,b: a[cmp] == b[cmp]
		elif isinstance(cmp, tuple):
			ca, cb = cmp
			cmp = lambda a ,b: a[ca] == b[cb]
		if merge_fn is None:
			def merge_fn(a ,b):
				a = a.copy()
				a.update(b)
				return a
		
		for a in self:
			for b in other:
				if cmp(a ,b):
					yield merge_fn(a ,b)
	
	def join_(self, other, cmp, merge_fn=None):
		if merge_fn is None:
			merge_fn = lambda a ,b: a.update(b)
		for _ in self._join(other, cmp, merge_fn=merge_fn):
			pass
	
	def join(self, other, cmp, merge_fn=None):
		return self.__class__(self._join(other, cmp, merge_fn=merge_fn))
	
	def filter_(self, fn):
		self.__dict__['_data'] = [x for x in self if fn(x)]
	
	def filter(self, fn):
		return self.__class__(x for x in self if fn(x))
	
	def new(self, *args, **kwargs):
		obj = self._el_type(*args, **kwargs)
		self.append(obj)
		return obj
	
	def map(self, fn, indexed=False, safe=False, pbar=None, reduce=None):
		'''
		fn is a callable taking one run as input
		'''
		
		outs = []
		
		seq = self if pbar is None else pbar(self)
		
		for i, x in enumerate(seq):
			try:
				inp = (i ,x) if indexed else (x,)
				out = fn(*inp)
				outs.append(out)
			except Exception as e:
				if safe:
					print(f'elm {i} failed')
					traceback.print_exc()
				else:
					raise e
		
		if pbar is not None:
			seq.close()
		
		if reduce is not None:
			return reduce(outs)
		return outs
	
	# def through(self, **map_kwargs):
	#
	# 	def _execute(fn, args=[], kwargs={}):
	# 		return self.map(lambda run: fn(run, *args, **kwargs),
	# 		                **map_kwargs)
	#
	# 	return make_ghost(self._el_type, _execute)

class _nothing:
	pass

class Key_Table(Table):
	'''
	Very similar to Table, but with a specified key that can be used with the get() function
	(so the table acts a little more like a dict)
	'''
	
	def __init__(self, *args, _get_key=None, **kwargs):
		super().__init__(*args, **kwargs)
		self.__dict__['_keys'] = {}
		self.__dict__['_get_key'] = _get_key
	
	def new(self, *args, **kwargs):
		obj = super().new(*args, **kwargs)
		get_key = self.__dict__['_get_key']
		self.__dict__['_keys'][get_key(obj)] = obj
		return obj

	def get(self, key, default=_nothing):
		if default is _nothing:
			return self.__dict__['_keys'][key]
		return self.__dict__['_keys'].get(key, default)




import sys, os
from pathlib import Path
import json
import yaml
import pickle

from . import unspecified_argument, agnosticmethod
from collections import OrderedDict

def monkey_patch(cls, module=None, include_mp=True):
	if module is None:
		import __main__
		module = __main__
		
		# try:
		# 	import __mp_main__
		# except ImportError:
		# 	pass
		# else:
		# 	monkey_patch(cls, module=__mp_main__)
	
	setattr(module, cls.__name__, cls)
	cls.__module__ = module.__name__

def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
	class OrderedLoader(Loader):
		pass
	def construct_mapping(loader, node):
		loader.flatten_mapping(node)
		return object_pairs_hook(loader.construct_pairs(node))
	OrderedLoader.add_constructor(
		yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
		construct_mapping)
	return yaml.load(stream, OrderedLoader)

# usage example:
# ordered_load(stream, yaml.SafeLoader)

def ordered_dump(data, stream=None, Dumper=yaml.Dumper, **kwds):
	class OrderedDumper(Dumper):
		pass
	def _dict_representer(dumper, data):
		return dumper.represent_mapping(
			yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
			data.items())
	OrderedDumper.add_representer(OrderedDict, _dict_representer)
	return yaml.dump(data, stream, OrderedDumper, **kwds)

# usage:
# ordered_dump(data, Dumper=yaml.SafeDumper)

def load_yaml(path, ordered=False):
	path = str(path)
	with open(path, 'r') as f:
		if ordered:
			return ordered_load(f, yaml.SafeLoader)
		return yaml.safe_load(f)

		# return yaml.load(f)

def save_yaml(data, path, ordered=False, default_flow_style=None, **kwargs):
	path = str(path)
	with open(path, 'w') as f:
		if ordered:
			return ordered_dump(data, stream=f, Dumper=yaml.SafeDumper,
			                    default_flow_style=default_flow_style, **kwargs)
		return yaml.safe_dump(data, f, default_flow_style=default_flow_style, **kwargs)


def load_json(path, **kwargs):
	path = str(path)
	with open(path, 'r') as f:
		return json.load(f, **kwargs)

def save_json(data, path, **kwargs):
	path = str(path)
	with open(path, 'w') as f:
		return json.dump(data, f, **kwargs)



def load_txt(path):
	path = str(path)
	with open(path, 'r') as f:
		return f.read()

def save_txt(data, path):
	path = str(path)
	with open(path, 'w') as f:
		return f.write(data)


def save_pickle(data, path):
	with open(str(path), 'wb') as f:
		pickle.dump(data, f)


def load_pickle(path):
	with open(str(path), 'rb') as f:
		return pickle.load(f)


def create_dir(directory):
	directory = str(directory)
	if not os.path.exists(directory):
		os.makedirs(directory)

def crawl(d, cond):
	if os.path.isdir(d):
		options = []
		for f in os.listdir(d):
			path = os.path.join(d, f)
			options.extend(crawl(path, cond))
		return options
	if cond(d):
		return [d]
	return []

def load_csv(path, sep=',', head=0, tail=0, row_sep='\n', as_gen=False):
	'''
	
	:param path:
	:param sep:
	:param head: number of lines to skip at the beginning of the file
	:param tail: number of lines to skip at the end of the file
	:return:
	'''

	with open(path, 'r') as f:
		raw = f.read()

	lines = raw.split(row_sep)
	if head > 0:
		lines = lines[min(head, len(lines)):]
	if tail > 0:
		lines = lines[:-min(tail, len(lines))]
		
	rows = (line.split(sep) for line in lines)
	
	if as_gen:
		return rows
	return list(rows)

def load_tsv(path, **kwargs):
	kwargs['sep'] = '\t'
	return load_csv(path, **kwargs)


def spawn_path_options(path):
	options = set()
	
	if os.path.isfile(path):
		options.add(path)
		path = os.path.dirname(path)
	
	if os.path.isdir(path):
		options.add(path)
	
	# TODO: include FIG_PATH_ROOT
	
	return options



class Persistent:
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self._datafiles = {}


	def _get_datafile_path(self, path, ext=unspecified_argument, quiet=False):
		'''if no name is provided, the head of path is used. if no ext is provided, it will try to be inferred
		by path contents, but no ext is used if it is ambiguous'''
		path = Path(path)
		root, name = path.parents[0], path.name #path.stem, path.suffix
		if ext is unspecified_argument and root.exists():
			options = list(root.glob(f'{name}*'))
			if len(options) == 1:
				name = options[0].name
		elif ext is not unspecified_argument and ext is not None:
			name = f'{name}.{ext}'
		if not quiet:
			create_dir(root)
			# root.mkdir(exist_ok=True)
		return root / name


	def _save_datafile(self, data, path, ext=unspecified_argument, overwrite=False, _save_fn=None):
		if ext is unspecified_argument:
			ext = 'pk'
		path = self._get_datafile_path(path, ext=ext)
		if _save_fn is None:
			_save_fn = save_pickle
		if not path.exists() or overwrite:
			_save_fn(data, str(path))
			return path


	def _load_datafile(self, path=None, name=None, ext=unspecified_argument, _load_fn=None):
		if ext is unspecified_argument:
			ext = 'pk'
		path = self._get_datafile_path(path=path, ext=ext)
		if _load_fn is None:
			_load_fn = load_pickle
		return _load_fn(str(path))


	def has_datafile(self, ident, root=None, ext=unspecified_argument, persistent=False):
		fixed, *suffix = ident.split('.')
		if ext is unspecified_argument and len(suffix):
			ext = '.'.join(suffix)
		if (ident in self._datafiles or fixed in self._datafiles) and not persistent:
			return True

		path = self._get_datafile_path(fixed if root is None else Path(root, fixed), ext=ext)
		return path.exists()


	def update_datafile(self, ident, data, root=None, overwrite=True, skip_save=False,
	                    ext=unspecified_argument, **kwargs):
		fixed, *suffix = ident.split('.')
		if ext is None and len(suffix):
			ext = '.'.join(suffix)
		path = fixed if root is None else Path(root, fixed)

		self._datafiles[fixed] = data
		if not skip_save:
			return self._save_datafile(data, path=path, ext=ext, overwrite=overwrite, **kwargs)


	def get_datafile(self, ident, root=None, ext=unspecified_argument, force_load=False, skip_cache=False,
	                 default=unspecified_argument, **kwargs):
		fixed, *suffix = ident.split('.')
		if ext is None and len(suffix):
			ext = '.'.join(suffix)
		path = fixed if root is None else Path(root, fixed)

		if not force_load:
			if ident in self._datafiles:
				return self._datafiles[ident]

			if fixed in self._datafiles:
				return self._datafiles[fixed]

		try:
			self._datafiles[fixed] = self._load_datafile(path, ext=ext, **kwargs)
		except FileNotFoundError:
			if default is unspecified_argument:
				raise
			self.update_datafile(fixed, default, root=root, ext=ext)
		out = self._datafiles[fixed]
		if skip_cache:
			del self._datafiles[fixed]
		return out


	def store_datafiles(self, datafiles=None, root=None, overwrite=False, ext=unspecified_argument, **kwargs):
		if datafiles is None:
			datafiles = self._datafiles
		return {name: self.update_datafile(name, value, root=root, overwrite=overwrite, ext=ext, **kwargs)
		        for name, value in datafiles.items()}



class HierarchyPersistent(Persistent):


	def _save_datafile(self, data, path, ext=unspecified_argument, overwrite=False,
	                  separate_dict=True, recursive=False, **kwargs):
		top = None if separate_dict and isinstance(data, dict) else ext
		path = self._get_datafile_path(path, ext=top)

		if top is None:
			path.mkdir(exist_ok=True)
			for key, value in data.items():
				self._save_datafile(value, path=path / key, overwrite=overwrite,
				                   separate_dict=separate_dict and recursive, recursive=recursive, **kwargs)
			return path
		return super()._save_datafile(data, path=path, ext=None, overwrite=overwrite, **kwargs)


	def _load_datafile(self, path, ext=unspecified_argument, delimiter='/', **kwargs):
		if isinstance(path, str) and delimiter is not None:
			path = Path(*path.split(delimiter))

		path = self._get_datafile_path(path=path, ext=ext)
		if path.is_dir():
			return {p.stem.split('.')[0]: self._load_datafile(path=p, delimiter=delimiter, **kwargs)
			        for p in path.glob('*')}
		return super()._load_datafile(path=path, ext=None, **kwargs)














from typing import NoReturn, Union, Any, Type



class Transactionable(object):
	'''
	Mixin to enable beginning, committing, and aborting transactions (multiple statements).
	To use Transactionable functionality, subclasses must implement begin(), in_transaction(), commit(), and abort().
	'''
	
	def begin(self) -> NoReturn:
		'''
		Must be overridden by subclasses.
		This prepares `self` to track all changes to `self` until commit() or abort() is called.
		If any attributes or data kept by `self`, this method should probably also call begin() in them.
		
		Has no effect if `self` is already in a transaction.
		
		:return: None
		'''
		raise NotImplementedError
	
	def in_transaction(self) -> bool:
		'''
		Query whether `self` is in a transaction.
		
		:return: True if and only if `self` is currently in a transaction.
		'''
		raise NotImplementedError
	
	def commit(self) -> NoReturn:
		'''
		Make all the changes to `self` since transaction began, and stop tracking changes from now on.
		
		Has no effect if `self` is not in a transaction.
		
		:return: None
		'''
		raise NotImplementedError
	
	def abort(self) -> NoReturn:
		'''
		Revert all changes to `self` since the transaction began, and stop tracking changes from now on.
		
		Has no effect if `self` is not in a transaction.
		
		:return: None
		'''
		raise NotImplementedError

	def __enter__(self) -> NoReturn:
		self.begin()
	
	def __exit__(self, type: Union[type(None), Type], *args: Any) -> Union[type(None), Type]:
		'''
		Once the context is ended, if no exception was raised the transaction is committed, otherwise,
		
		:param type: Either None or an Exception type, if an Exception was raised in the context.
		:param args: Other possible args provided by the raised Exception
		:return: If there was no exception, or the
		'''
		if type is None:
			self.commit()
		else:
			self.abort()
		return None if type is None else type.__name__ == 'AbortTransaction'
	

class AbortTransaction(Exception):
	pass



from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Type, Iterator, Iterable
from datetime import datetime, timezone
from collections import OrderedDict, UserList, UserDict
from .ordered_set import OrderedSet
from .typing import agnosticmethod, unspecified_argument


class OmniStructure:
	pass



class OmniNode:
	def __copy__(self, **kwargs):
		return self.__class__(**kwargs)



class IntrinsicStructure(OmniStructure):
	'''The structure is fully defined by the collection of nodes (eg. linked list)'''
	def register(self, node: OmniNode) -> OmniStructure:
		raise NotImplementedError
	
	
	def deregister(self, node: OmniNode) -> OmniNode:
		raise NotImplementedError



class ExtrinsicStructure(OmniStructure):
	'''The structure includes auxillary information beyond the nodes (eg. dict)'''
	def register(self, addr: Hashable, node: OmniNode) -> OmniStructure:
		raise NotImplementedError
	
	
	def deregister(self, addr: Hashable) -> OmniNode:
		raise NotImplementedError



class ContextStructure(OmniStructure): # abstract
	_structure_instance_for_context = None
	def __new__(cls, *args, **kwargs):
		if cls._structure_instance_for_context is None:
			return super().__new__(cls, *args, **kwargs)
		return cls._structure_instance_for_context
	
	
	def __enter__(self):
		self.__class__._structure_instance_for_context = self
	
	
	def __exit__(self, exc_type, exc_val, exc_tb):
		del self.__class__._structure_instance_for_context



class PayloadNode(OmniNode):
	@property
	def payload(self) -> Any:
		raise NotImplementedError



class StructureNode(OmniNode):
	def __init__(self, structure: OmniStructure = None, **kwargs):
		super().__init__(**kwargs)
		self._structure = None
		self._structure = self._create_structure(structure)


	@property
	def structure(self) -> OmniStructure:
		return self._structure
	
	
	Structure = None
	def _create_structure(self, structure: OmniStructure = None, **kwargs) -> OmniStructure:
		if structure is None:
			structure = self.Structure(**kwargs)
		return structure.register(self)


	def _deregister(self):
		raise NotImplementedError

	
	def __copy__(self, structure: Optional[OmniStructure] = unspecified_argument, **kwargs):
		if structure is unspecified_argument:
			structure = self._structure
		return super().__copy__(structure=structure, **kwargs)
		
	
	def isolated_copy(self, structure=None, **kwargs):
		return self.__copy__(structure=None, **kwargs)
	


class AssertiveIntrinsicStructure(IntrinsicStructure):
	def register(self, node: StructureNode) -> OmniStructure:
		if node.structure is not self:
			node._structure = self
		return super().register(node)
		


class IdentNode(OmniNode):
	@property
	def identifier(self) -> Hashable:
		return hex(id(self))[2:]
	
	
	def __str__(self):
		return self.identifier
	
	
	def __repr__(self):
		return f'{self.__class__.__name__}:{self.identifier}'
	
	
	def __eq__(self, other):
		return self.identifier == other.identifier
	
	
	def __hash__(self):
		return hash(self.identifier)



class ExtrinsicToIntrinsicStructure(ExtrinsicStructure, IntrinsicStructure):
	def register(self, node: IdentNode) -> OmniStructure:
		return super().register(node.identifier, node)


	def deregister(self, node: IdentNode) -> OmniNode:
		return super().deregister(node.identifier)



class SequenceStructure(IntrinsicStructure, OrderedSet):
	def register(self, node: OmniNode) -> 'SequenceStructure':
		self.add(node)
		return self
	
	
	def deregister(self, node: OmniNode) -> 'SequenceStructure':
		self.remove(node)
		return self



class TableStructure(ExtrinsicStructure, OrderedDict):
	def register(self, addr: Hashable, node: OmniNode) -> 'TableStructure':
		self[addr] = node
		return self
	
	
	def deregister(self, addr: Hashable) -> OmniNode:
		node = self[addr]
		del self[addr]
		return node
	
	
	
class BaseStructure(OmniStructure):
	def __init__(self, base: OmniNode, structure: OmniStructure = None):
		super().__init__()
		self.base = base
		self.structure = structure
	
	
	def deregister_base(self):
		pass



class NodeEdgeStructure(IntrinsicStructure, OrderedDict):
	class NodeEdges(BaseStructure):
		pass


	def register(self, node: IdentNode, **kwargs) -> NodeEdges:
		ID = node.identifier
		if ID in self:
			return self[ID]
		edges = self.NodeEdges(node, self, **kwargs)
		self[ID] = edges
		return edges


	def deregister(self, node: IdentNode) -> IdentNode:
		ID = node.identifier
		self[ID].deregister_base()
		del self[ID]
		return node



class AbstractParentNode(OmniNode):
	@property
	def parent(self) -> OmniNode:
		raise NotImplementedError


	@property
	def children(self) -> OmniStructure:
		raise NotImplementedError



class ParentNode(AbstractParentNode): # single parent, many children
	def __init__(self,
	             parent: Optional[AbstractParentNode] = None,
	             children: Optional[Sequence[AbstractParentNode]] = None,
	             **kwargs):
		super().__init__(**kwargs)
		self._parent: Union[ParentNode, None] = None
		self.parent = parent
		if children is not None:
			self.add_children(*children)
	
	
	@property
	def parent(self) -> 'ParentNode':
		return self._parent
	@parent.setter
	def parent(self, parent: 'ParentNode'):
		if self._parent is not None:
			self.parent.children.deregister(self)
		if parent is not None:
			parent.children.register(self)
		self._parent = parent


	@property
	def children(self) -> IntrinsicStructure:
		raise NotImplementedError
	
	
	def add_children(self, *children: OmniNode) -> 'ParentNode':
		registry = self.children
		for child in children:
			registry.register(child)
		return self



class AbstractMultiParentNode(OmniNode):
	@property
	def parents(self) -> OmniStructure:
		raise NotImplementedError


	@property
	def children(self) -> OmniStructure:
		raise NotImplementedError



class MultiParentNode(AbstractMultiParentNode):  # single parent, many children
	def __init__(self,
	             parents: Optional[Sequence[AbstractMultiParentNode]] = None,
	             children: Optional[Sequence[AbstractMultiParentNode]] = None,
	             **kwargs):
		super().__init__(**kwargs)
		if parents is not None:
			self.add_parents(*parents)
		if children is not None:
			self.add_children(*children)
	

	def add_parents(self, *parents: 'MultiParentNode') -> 'MultiParentNode':
		registry = self.parents
		for parent in parents:
			registry.register(parent)
			# parent.children.register(self)
		return self
	
	
	def add_children(self, *children: 'MultiParentNode') -> 'MultiParentNode':
		registry = self.children
		for child in children:
			registry.register(child)
			# child.parents.register(self)
		return self



class GraphNode(StructureNode):
	class Structure(NodeEdgeStructure):
		class NodeEdges(NodeEdgeStructure.NodeEdges, SequenceStructure):
			def register(self, node: StructureNode) -> 'NodeEdges':
				node.structure.add(self.base)
				return super().register(node)
			
			
			def deregister(self, node: StructureNode) -> ParentNode:
				node = super().deregister(node)
				node.structure.remove(self.base)
				return node
			
			
			def deregister_base(self):
				for node in self:
					node.structure.remove(self.base)
				return super().deregister_base()



class GlobalTreeNode(ParentNode, StructureNode):
	@property
	def children(self) -> OmniStructure:
		return self.structure
	
	
	class Structure(NodeEdgeStructure):
		class NodeEdges(NodeEdgeStructure.NodeEdges):
			def register(self, node: ParentNode) -> 'NodeEdges':
				node._parent = self.base
				return super().register(node)
			
			
			def deregister(self, node: ParentNode) -> ParentNode:
				node = super().deregister(node)
				node._parent = None
				return node
			
			
			def deregister_base(self):
				for node in self:
					node._parent = None


class TableTreeNode(StructureNode): # Table Tree Node (tree where children have an address)
	class Structure(NodeEdgeStructure):
		class NodeEdges(NodeEdgeStructure.NodeEdges, TableStructure):
			def register(self, addr: Hashable, node: ParentNode) -> 'NodeEdges':
				node._parent = self.base
				return super().register(addr, node)

			def deregister(self, addr: Hashable) -> ParentNode:
				node = super().deregister(addr)
				node._parent = None
				return node

			def deregister_base(self):
				for node in self:
					node._parent = None



class DiGraphNode(MultiParentNode, StructureNode):
	class Structure(NodeEdgeStructure):
		class Parents(BaseStructure, SequenceStructure):
			def register(self, node: 'DiGraphNode') -> 'NodeEdges':
				node.children.add(self.base)
				return super().register(node)
			
			def deregister(self, node: 'DiGraphNode') -> 'DiGraphNode':
				node = super().deregister(node)
				node.children.remove(self.base)
				return node
			
			def deregister_base(self):
				for node in self:
					node.children.remove(self.base)
				return super().deregister_base()
		
		
		class Children(BaseStructure, SequenceStructure):
			def register(self, node: 'DiGraphNode') -> 'NodeEdges':
				node.parents.add(self.base)
				return super().register(node)
			
			def deregister(self, node: 'DiGraphNode') -> 'DiGraphNode':
				node = super().deregister(node)
				node.parents.remove(self.base)
				return node
			
			def deregister_base(self):
				for node in self:
					node.parents.remove(self.base)
				return super().deregister_base()
		
		
		class NodeEdges(NodeEdgeStructure.NodeEdges):
			def __init__(self, base: OmniNode, structure: OmniStructure,
			             parents: 'Parents', children: 'Children'):
				super().__init__(base, structure)
				self._parents = parents
				self._children = children
			
			
			@property
			def parents(self) -> 'Parents':
				return self._parents
			
			
			@property
			def children(self) -> 'Children':
				return self._children


			def deregister_base(self):
				for parent in self.parents:
					parent.children.deregister(self.base)
				for child in self.children:
					child.parents.deregister(self.base)
		
		
		def register(self, node: IdentNode, parents=unspecified_argument, children=unspecified_argument,
		             **kwargs) -> NodeEdges:
			if parents is unspecified_argument:
				parents = self.Parents(node, self)
			if children is unspecified_argument:
				children = self.Children(node, self)
			return super().register(node, parents=parents, children=children, **kwargs)

	
	@property
	def parents(self) -> 'Structure.Parents':
		return self.structure.parents

	
	@property
	def children(self) -> 'Structure.Children':
		return self.structure.children



class StampedNode(OmniNode):
	def __init__(self, timestamp=unspecified_argument, **kwargs):
		super().__init__(**kwargs)
		self._timestamp = datetime.now(tz=timezone.utc) if timestamp is unspecified_argument else timestamp
	
	
	def copy(self, timestamp=unspecified_argument, **kwargs):
		if timestamp is unspecified_argument:
			timestamp = self.timestamp
		return self.__copy__(timestamp=timestamp, **kwargs)
	
	
	@property
	def timestamp(self):
		return self._timestamp
	
	
	# def __hash__(self):
	# 	return hash(self.timestamp)
	#
	# def __eq__(self, other):
	# 	return self.timestamp == other.timestamp
	
	
	def __lt__(self, other):
		return self.timestamp < other.timestamp
	
	
	def __le__(self, other):
		return self.timestamp <= other.timestamp
	
	
	def __gt__(self, other):
		return self.timestamp > other.timestamp
	
	
	def __ge__(self, other):
		return self.timestamp >= other.timestamp



####################################################################################################

# if you need global structure in one object see TreeNode and TableTreeNode above (children, but not parent).
# if not, use LocalNode (parent and children only stored locally)


class LocalNode(PayloadNode):
	# DefaultNode = None
	ChildrenStructure = None

	class empty_value: pass
	empty_value.payload = empty_value

	class _MissingKey(KeyError): pass

	@classmethod
	def from_raw(cls, raw: Any, *, parent: Optional['LocalNode'] = unspecified_argument,
	             parent_key: Optional[str] = None, **kwargs) \
			-> Union['LocalNode', empty_value]:
		if isinstance(raw, LocalNode):
			raw.parent = parent
			raw._parent_key = parent_key
			return raw
		if raw is cls.empty_value or isinstance(raw, cls.empty_value):
			return raw
		return cls(payload=raw, parent=parent, **kwargs)

	def my_address(self):
		parent = self.parent
		return () if parent is None else parent.my_address() + (self.parent_key,)

	def __init__(self, payload=unspecified_argument, *, parent: Optional['LocalNode'] = unspecified_argument,
	             children: Optional[ChildrenStructure] = unspecified_argument,
	             parent_key: Optional[str] = None, **kwargs):
		super().__init__(**kwargs)
		self._payload = payload
		self._parent = parent
		if children is unspecified_argument:
			children = self.ChildrenStructure()
		self._children = children
		self._parent_key = parent_key


	@property
	def root(self) -> 'LocalNode':
		parent = self.parent
		if parent is None:
			return self
		return parent.root
	
	@property
	def parent(self):
		if self.has_parent:
			return self._parent
	@parent.setter
	def parent(self, value):
		self._parent = value
	@property
	def has_parent(self):
		return self._parent is not unspecified_argument
	@property
	def parent_key(self):
		return self._parent_key

	
	@property
	def payload(self):
		if self.has_payload:
			return self._payload
		return self._default_payload()
	@payload.setter
	def payload(self, value):
		self._payload = value
	@property
	def has_payload(self):
		return self._payload is not unspecified_argument


	@property
	def is_leaf(self):
		for _ in self.children(skip_empty=True):
			return False
		return True

	def __len__(self):
		return self._num_children()

	def __contains__(self, addr: Hashable):
		return self.has(addr)

	def __getitem__(self, addr: Hashable):
		return self.get(addr)

	def __setitem__(self, addr: Hashable, node: 'LocalNode'):
		return self.set(addr, node)

	def __delitem__(self, addr: Hashable):
		return self.remove(addr)

	def __iter__(self):
		return self.children()


	def _num_children(self):
		raise NotImplementedError


	def _get(self, addr: Hashable):
		raise NotImplementedError


	def _set(self, addr: Hashable, node: 'LocalNode'):
		raise NotImplementedError


	def _remove(self, key):
		raise NotImplementedError


	def _has(self, addr: Hashable):
		raise NotImplementedError


	def _iterate_children(self):
		raise NotImplementedError


	def _default_payload(self):
		return self.to_python()


	def to_python(self):
		if not self.has_payload:
			raise ValueError('no payload')
		return self.payload


	def named_children(self, skip_empty=True):
		for key, child in self._iterate_children():
			if skip_empty and child.is_leaf and not child.has_payload:
				continue
			yield key, child

	def children(self, skip_empty=True):
		for key, child in self.named_children(skip_empty=skip_empty):
			yield child

	def add_all(self, children: Iterable[Tuple[Hashable, 'LocalNode']], **kwargs) -> None:
		for addr, node in children:
			self.set(addr, node, **kwargs)


	def get(self, addr: Hashable, default: Any = unspecified_argument) -> 'LocalNode':
		try:
			return self._get(addr)
		except self._MissingKey:
			if default is unspecified_argument:
				raise
			return default


	def set(self, addr: Hashable, value: Any, **kwargs) -> 'LocalNode':
		node = self.from_raw(value, parent=self, parent_key=addr, **kwargs)
		self._set(addr, node)
		return node


	def has(self, addr: Hashable) -> bool:
		return self._has(addr)


	def remove(self, addr: Hashable) -> None:
		return self._remove(addr)



class SparseNode(LocalNode):
	ChildrenStructure = OrderedDict
	_python_structure = OrderedDict

	def _get(self, addr: Hashable) -> LocalNode:
		try:
			return self._children[addr]
		except KeyError:
			raise self._MissingKey(addr)

	def _set(self, addr: Hashable, node: LocalNode):
		self._children[addr] = node

	def _remove(self, addr: Hashable):
		del self._children[addr]

	def _has(self, addr: Hashable):
		return addr in self._children

	def _num_children(self):
		return len(self._children)

	def _iterate_children(self):
		return self._children.items()

	def to_python(self):
		return self._python_structure([(key, value.payload) for key, value in self._children.items()])



class DenseNode(LocalNode):
	ChildrenStructure = list
	_python_structure = list

	def _parse_index(self, addr: Hashable, strict: bool = False) -> int:
		if isinstance(addr, str):
			# if addr == '_':
			# 	addr = len(self)
			# else:
			addr = int(addr)
		if not isinstance(addr, int):
			raise ValueError(addr)
		if strict and not (-len(self._children) <= addr < len(self._children)):
			raise IndexError(addr)
		return addr

	def _get(self, addr: Hashable) -> LocalNode:
		try:
			return self._children[self._parse_index(addr, strict=True)]
		except (IndexError, ValueError):
			raise self._MissingKey(addr)

	def _set(self, addr: Hashable, node: LocalNode):
		idx = self._parse_index(addr, strict=False)
		if idx == len(self._children):
			self._children.append(node)
		elif -len(self._children) <= idx < len(self._children):
			self._children[idx] = node
		else:
			raise IndexError(addr)

	def _remove(self, addr: Hashable):
		del self._children[self._parse_index(addr, strict=True)]

	def _has(self, addr: Hashable):
		try:
			idx = self._parse_index(addr, strict=False)
		except IndexError:
			return False
		except ValueError:
			return False
		return -len(self._children) <= idx < len(self._children)

	def _num_children(self):
		return len(self._children)

	def _iterate_children(self):
		for i, x in enumerate(self._children):
			yield str(i), x

	def to_python(self):
		return self._python_structure(value.payload for value in self._children)

	def prepend(self, val: Any):
		self._children.insert(0, self.from_raw(val))

	def append(self, val: Any):
		self._children.append(self.from_raw(val))

	def insert(self, idx: int, val: Any):
		self._children.insert(idx, self.from_raw(val))

	def extend(self, vals: Iterable[Any]):
		self._children.extend(self.from_raw(val) for val in vals)



class TreeNode(LocalNode):
	DefaultNode: Type['TreeNode'] = SparseNode
	DenseNode: Type['TreeNode'] = None
	SparseNode: Type['TreeNode'] = None


	# def __new__(cls, raw: Any = unspecified_argument, **kwargs):
	# 	return cls.from_raw(raw)

	
	@classmethod
	def from_raw(cls, raw: Any, *, parent: Optional['LocalNode'] = unspecified_argument,
	             parent_key: Optional[str] = None, **kwargs) -> 'LocalNode':
		if isinstance(raw, LocalNode):
			raw.parent = parent
			raw._parent_key = parent_key
			return raw
		if isinstance(raw, dict):
			node = cls.SparseNode(parent=parent, parent_key=parent_key, **kwargs)
			for key, value in raw.items():
				node.set(key, cls.from_raw(value, parent=node, **kwargs), **kwargs)
		elif isinstance(raw, (tuple, list)):
			node = cls.DenseNode(parent=parent, parent_key=parent_key, **kwargs)
			for idx, value in enumerate(raw):
				idx = str(idx)
				node.set(idx, cls.from_raw(value, parent=node, **kwargs), **kwargs)
		else:
			node = cls.DefaultNode(payload=raw, parent=parent, **kwargs)
		return node


	@classmethod
	def from_dict(cls, raw: Dict[str, Any], *, parent: Optional['LocalNode'] = unspecified_argument,
	              **kwargs) -> 'LocalNode':
		return cls.from_raw(raw, parent=parent, **kwargs)



class TreeSparseNode(SparseNode, TreeNode): pass
class TreeDenseNode(DenseNode, TreeNode): pass

TreeNode.DefaultNode = TreeSparseNode
TreeNode.SparseNode = TreeSparseNode
TreeNode.DenseNode = TreeDenseNode



class ConvertableNode(TreeNode):
	def _convert_child(self, addr: Hashable, new_type: Type[LocalNode]) -> LocalNode:
		old = self.get(addr)
		new = new_type(parent=self)
		for key, value in old.children():
			new.set(key, value)
		self.set(addr, new)
		return new
	
	
	def convert_to_sparse(self, addr: Hashable) -> LocalNode:
		return self._convert_child(addr, self.SparseNode)


	def convert_to_dense(self, addr: Hashable) -> LocalNode:
		return self._convert_child(addr, self.DenseNode)



class AddressNode(LocalNode):
	_address_delimiter = '.'
	
	
	class _ConnectorError(LocalNode._MissingKey):
		def __init__(self, node, current, rest):
			super().__init__(current)
			self.node = node
			self.current = current
			self.rest = rest
	
	
	def _evaluate_address(self, addr: str) -> Tuple['AddressNode', str]:
		current, *rest = str(addr).split(self._address_delimiter)
		if len(rest):
			try:
				node = self[current] if len(current) else self.parent
			except KeyError:
				node = None
			if not isinstance(node, AddressNode):
				raise self._ConnectorError(self, current, rest)
			return node._evaluate_address(self._address_delimiter.join(rest))
		return self, current


	def flatten(self, include_connector_payloads=True, skip_empty=True) -> Iterator[Tuple[str, 'LocalNode']]:
		for key, value in self.children(skip_empty=skip_empty):
			assert isinstance(value, AddressNode), f'Unexpected node type: {type(value)}'
			if (include_connector_payloads and value.has_payload) or value.is_leaf:
				yield key, value.payload
			for subkey, subvalue in value.flatten(include_connector_payloads=include_connector_payloads,
			                                      skip_empty=skip_empty):
				yield f'{key}{self._address_delimiter}{subkey}', subvalue

	def get(self, addr: str, default: Any = unspecified_argument) -> 'LocalNode':
		node, key = self._evaluate_address(addr)
		return super(AddressNode, node).get(key, default)
		

	def set(self, addr: str, value: Any, **kwargs) -> 'LocalNode':
		node, key = self._evaluate_address(addr)
		return super(AddressNode, node).set(key, value, **kwargs)


	def has(self, addr: str) -> bool:
		try:
			node, key = self._evaluate_address(addr)
		except self._ConnectorError:
			return False
		return super(AddressNode, node).has(key)


	def remove(self, addr: str) -> None:
		node, key = self._evaluate_address(addr)
		return super(AddressNode, node).remove(key)



class AutoAddressNode(AddressNode):
	def _auto_create_child(self, key):
		return self.set(key, self.__class__(parent=self))

	def _evaluate_address(self, addr: str, auto_create: bool = False) -> Tuple['AddressNode', str]:
		try:
			return super()._evaluate_address(addr)
		except self._ConnectorError as e:
			if not auto_create:
				raise
			e.node._auto_create_child(e.current)
			return e.node._evaluate_address(self._address_delimiter.join((e.current, *e.rest)),
			                                auto_create=auto_create)

	def set(self, addr: str, value: Any, **kwargs):
		node, key = self._evaluate_address(addr, auto_create=True)
		return super(AddressNode, node).set(key, value, **kwargs)



class AutoTreeNode(AutoAddressNode, ConvertableNode, TreeNode):
	def _auto_create_child(self, key):
		node = self.DenseNode(parent=self) if key.isdigit() else self.SparseNode(parent=self)
		self.set(key, node)
class AutoTreeSparseNode(SparseNode, AutoTreeNode): pass
class AutoTreeDenseNode(DenseNode, AutoTreeNode):
	def _has(self, addr: Hashable):
		try:
			idx = self._parse_index(addr, strict=False)
		except IndexError:
			pass
		except TypeError:
			pass
		else:
			if -len(self._children) <= idx < len(self._children):
				return self._children[idx] is not self.empty_value
		return False

	def _set(self, addr: Hashable, node: LocalNode):
		idx = self._parse_index(addr, strict=False)
		if idx == len(self._children):
			self._children.append(node)
		elif -len(self._children) <= idx < len(self._children):
			self._children[idx] = node
		elif idx > 0:
			self._children.extend([self.empty_value] * (idx - len(self._children)))
			self._children.append(node)
		else:
			raise IndexError(idx)

AutoAddressNode.DefaultNode = AutoTreeSparseNode
AutoTreeNode.SparseNode = AutoTreeSparseNode
AutoTreeNode.DenseNode = AutoTreeDenseNode


####################################################################################################


class LeafNode(PayloadNode):
	def __init__(self, payload=unspecified_argument, **kwargs):
		super().__init__(**kwargs)
		self._payload = payload

	
	@property
	def payload(self):
		return self._payload



class TimelineNode(StructureNode):
	Structure = SequenceStructure
	# class Structure(SequenceStructure):
	# 	@property
	# 	def origin(self) -> StructureNode:
	# 		return self[0]
	
	
	@property
	def origin(self) -> StructureNode:
		return self.structure[0]
	
	
	def trajectory(self):
		idx = self.structure.index(self)
		for i in range(0, idx):
			yield self.structure[i]
	
	
	def past(self):
		idx = self.structure.index(self)
		for i in range(idx-1, -1, -1):
			yield self.structure[i]
	
	
	def future(self):
		idx = self.structure.index(self)
		for i in range(idx + 1, len(self.structure)):
			yield self.structure[i]
	



# class ConfigNode(Node):
# 	AntiNode = AntiNode
# 	ReferenceNode = ReferenceNode


#
# class SpaceTimeNode:
# 	def past(self, addr=None):
# 		raise NotImplementedError
#
# 	def future(self, addr=None):
# 		raise NotImplementedError
#
# 	def super(self, addr=None):
# 		raise NotImplementedError
#
# 	def sub(self, addr=None):
# 		raise NotImplementedError
	


# class AddressNode(SpaceTimeNode):
# 	def __init__(self, **kwargs):
# 		super().__init__(**kwargs)
# 		self._past_nodes = self.AddressBook()
# 		self._future_nodes = self.AddressBook()
# 		self._sub_nodes = self.AddressBook()
# 		self._super_nodes = self.AddressBook()
#
#
# 	class AddressBook:
# 		def __init__(self, **kwargs):
# 			super().__init__(**kwargs)
# 			self._addresses = []
# 			self._address_index = {}
#
# 		def cut(self, addr):
# 			if addr in self._address_index:
# 				index = self._address_index[addr]
# 				del self._addresses[index]
# 				del self._address_index[addr]
#
# 		def push(self, addr):
# 			if addr not in self._address_index:
# 				self._addresses.append(addr)
# 				self._address_index[addr] = len(self._addresses) - 1
#
#
# 		def recover(self, addr):
# 			if addr in self._address_index:
# 				return self._addresses[self._address_index[addr]]
#
#
# 		def next(self, addr=None):
# 			if addr is None and len(self._addresses) > 0:
# 				return self._addresses[0]
# 			if addr not in self._address_index or self._address_index[addr] == len(self._addresses) - 1:
# 				return None
# 			return self._addresses[self._address_index[addr] + 1]
#
#
# 		def search(self, addr=None):
# 			current = 0 if addr is None or addr not in self._address_index else self._address_index[addr]+1
# 			while current < len(self._addresses):
# 				yield self._addresses[current]
# 				current += 1
#
#
# 		def __len__(self):
# 			return len(self._addresses)
#
#
# 		def __contains__(self, item):
# 			return item in self._address_index
#
#
# 	def past(self, addr=None):
# 		return self._past_nodes.next(addr)
#
#
# 	def future(self, addr=None):
# 		return self._future_nodes.next(addr)
#
#
# 	def sub(self, addr=None):
# 		return self._sub_nodes.next(addr)
#
#
# 	def super(self, addr=None):
# 		return self._super_nodes.next(addr)
#
#
#
# class DataNode(SpaceTimeNode):
# 	def __init__(self, payload=None, **kwargs):
# 		super().__init__(**kwargs)
# 		self.data = payload
#
#
#
# class ReferenceNode(AddressNode, DataNode):
# 	pass
#
#
#
# class LeafNode(AddressNode, DataNode):
# 	pass
#
#
# class Structure:
# 	def update(self, addr=None, **kwargs):
# 		raise NotImplementedError
#
# 	def attach(self, addr=None, **kwargs):
# 		raise NotImplementedError
#
#
# 	def assemble(self, base=None, **kwargs):
# 		raise NotImplementedError
#
# 	def forecast(self, base=None, **kwargs):
# 		raise NotImplementedError
#
# 	def external(self, addr=None, **kwargs):
# 		raise NotImplementedError
#
# 	def internal(self, addr=None, **kwargs):
# 		raise NotImplementedError
#
# 	def owners(self, base=None, **kwargs):
# 		raise NotImplementedError
#
# 	def followers(self, base=None, **kwargs):
# 		raise NotImplementedError
#
#
#
#
# class BranchNode(AddressNode, DataNode):
#
# 	class _missing_value:
# 		pass
#
#
# 	class PullError(KeyError):
# 		def __init__(self, addr):
# 			super().__init__(str(addr))
# 			self.addr = addr
#
#
# 	def attach(self, node):
# 		self._sub_nodes.push(node)
# 		node._super_nodes.push(self)
#
#
# 	def update(self, node):
# 		self._future_nodes.push(node)
# 		node._past_nodes.push(self)
#
#
# 	def nodify(self, obj):
# 		'''Returns a node from the given object.'''
# 		raise NotImplementedError
#
#
# 	def pull(self, *addrs, default=unspecified_argument, **construct_args):
# 		trace = self._trace_pull(*addrs, default=default)
# 		return self.construct(trace, **construct_args)
#
#
# 	def pull_node(self, *addrs, default=unspecified_argument):
# 		trace = self._trace_pull(*addrs, default=default)
# 		return trace[-1]
# 		raise NotImplementedError
#
#
# 	class AddressTrace:
# 		def __init__(self, addrs=None, default=None, **kwargs):
# 			super().__init__(**kwargs)
# 			self.addrs = addrs
# 			self.default = default
# 			self.records = None
#
#
# 		class TracedRecord:
# 			__slots__ = ('trace', 'predecessor', 'record')
# 			def __init__(self, trace, predecessor=None, **kwargs):
# 				super().__init__(**kwargs)
# 				self.trace = trace
# 				self.predecessor = predecessor
# 				self.record = None
#
# 			def collect(self):
# 				records = [self.record]
# 				record = self
# 				while record.predecessor is not None:
# 					records.append(record.record)
# 					record = record.predecessor
# 				self.trace.records = records[::-1]
#
# 			def append(self, *args, **kwargs):
# 				self.record = args, kwargs
# 				return self.trace.create_record(predecessor=self, **kwargs)
#
#
# 		def create_record(self, predecessor=None, trace=None, **kwargs):
# 			if trace is None:
# 				trace = self
# 			return self.TracedRecord(trace=trace, predecessor=predecessor, **kwargs)
#
#
# 		def append(self, *args, **kwargs):
# 			return self.create_record(predecessor=self).append(*args, **kwargs)
#
#
# 		@classmethod
# 		def _flatten_trace(cls, path, node):
# 			'''Flattens the given trace.'''
# 			trace = [node]
# 			while len(path) > 1:
# 				trace.append(path[-1])
# 				path = path[0]
# 			trace.append(path[0])
# 			return trace[::-1]
#
#
# 	def _trace_pull(self, *addrs, default=unspecified_argument, trace=None):
# 		if default is not unspecified_argument:
# 			default = self.nodify(default)
# 		if trace is None:
# 			trace = self.AddressTrace(addrs=addrs, default=default)
# 		for addr in addrs:
# 			try:
# 				result = self._pull_remote(addr, trace=trace.append(addr=addr, node=self))
# 			except self.PullError:
# 				pass
# 			else:
# 				return result.collect()
# 		if default is not unspecified_argument:
# 			return trace
# 		raise self.PullError(addrs)
#
#
#
# 	def _pull_remote(self, addr, trace=None):
# 		try:
# 			return self._pull_local(addr, trace=trace)
# 		except self.PullError:
# 			# for past in self._past_nodes.search():
# 			# 	try:
# 			# 		return past._pull_remote(addr, trace=None if trace is None else trace.append(up=past))
# 			# 	except self.PullError:
# 			# 		pass
# 			for sup in self._super_nodes.search():
# 				try:
# 					return sup._pull_remote(addr, trace=None if trace is None else trace.append(left=sup))
# 				except self.PullError:
# 					pass
# 		raise self.PullError(addr)
#
#
# 	@staticmethod
# 	def _parse_addr(addr):
# 		if addr is None:
# 			return None
# 		if isinstance(addr, tuple):
# 			return addr
# 		if isinstance(addr, list):
# 			return tuple(addr)
# 		if isinstance(addr, str):
# 			if len(addr) == 0:
# 				return None
# 			return tuple(addr.split('.'))
# 		return (addr,)
#
#
# 	def _pull_local(self, addr, trace=None):
#
# 		terms = self._parse_addr(addr)
# 		if trace is not None and terms != addr:
# 			trace = trace.append(terms=terms, addr=addr)
# 		try:
# 			if terms is None or len(terms) == 0:
# 				return trace.append(terms=terms)
#
# 			if terms[0] == '':
# 				if len(terms) > 1:
# 					return self.super()._pull_local(terms[1:], trace=None if trace is None else trace.append(left=sup))
# 				return
# 			sub = self._sub_nodes.recover(terms[0])
# 			trace = (_trace, sub)
# 			if len(terms) > 1:
# 				return sub._pull_local(terms[1:], _trace=trace)
# 			return trace
# 		except self.PullError:
# 			for past in self._past_nodes.search():
# 				try:
# 					return past._pull_local(addr, _trace=trace)
# 				except self.PullError:
# 					pass
# 		raise self.PullError(addr)
#
#
# 	def push(self, addr, val):
# 		raise NotImplementedError
#
#
# 	def construct(self, trace, **kwargs):
# 		'''Processes and returns the payload of the node.'''
# 		return self.data
#
# 	pass
#
#
# class Integral:
# 	def touch(self, node):
# 		pass
#
# 	def mark(self, node):
# 		pass
#
# 	def untouched(self, nodes):
# 		pass
#
#
# class Trajectory:
# 	def step(self, current, target):
# 		pass
#
#
#
#
# class Node:
# 	def __init__(self, *args, **kwargs):
# 		super().__init__(*args, **kwargs)
# 		self._up = {}
# 		self._down = {}
# 		self._left = {}
# 		self._right = {}
#
#
# 	def _move_along_edge(self, direction, edges, key):
# 		if edges is None:
# 			return None
# 		return edges.get(key)
#
#
# 	def _set_single_edge(self, direction, edges, key, val=None):
# 		edges[key] = val
# 		return edges
#
#
# 	def up(self, key=None):
# 		return self._move_along_edge('up', edges=self._up, key=key)
#
#
# 	def down(self, key=None):
# 		return self._move_along_edge('down', edges=self._down, key=key)
#
#
# 	def left(self, key=None):
# 		return self._move_along_edge('left', edges=self._left, key=key)
#
#
# 	def right(self, key=None):
# 		return self._move_along_edge('right', edges=self._right, key=key)
#
#
# 	def set_up(self, key, val=None):
# 		self._set_single_edge('up', edges=self._up, key=key, val=val)
#
#
# 	def set_down(self, key, val=None):
# 		self._set_single_edge('down', edges=self._down, key=key, val=val)
#
#
# 	def set_left(self, key, val=None):
# 		self._set_single_edge('left', edges=self._left, key=key, val=val)
#
#
# 	def set_right(self, key, val=None):
# 		self._set_single_edge('right', edges=self._right, key=key, val=val)
#
#
#
# class DataNode(Node):
# 	def __init__(self, payload=None, **kwargs):
# 		super().__init__(**kwargs)
# 		self._payload = payload
#
#
# 	@property
# 	def payload(self):
# 		return self._payload










